/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef AE_QUEUE_HPP
#define AE_QUEUE_HPP

#include <pthread.h>
#include <unistd.h>

#include <cassert>
#include <queue>
#include <set>
#include <string>
#include <tr1/memory>
#include <stdint.h>

#include "segment.hpp"

struct AE_connection {
	AE_connection(int socket) :
		socket(socket)
	{}
	~AE_connection()
	{
		close(socket);
	}

	int		socket;

private:
	// non-copyable
	AE_connection(const AE_connection &) {}
	void operator=(const AE_connection &) {}
};

// a synchronized queue for client requests
class AE_queue {
public:
	AE_queue()
	{
		pthread_cond_init(&cond, 0);
		pthread_mutex_init(&mutex, 0);
	}
	~AE_queue()
	{
		pthread_cond_destroy(&cond);
		pthread_mutex_destroy(&mutex);
	}

	void push(std::tr1::shared_ptr<AE_connection> cx)
	{
		pthread_mutex_lock(&mutex);
		queue.push(cx);
		pthread_cond_signal(&cond);
		pthread_mutex_unlock(&mutex);
	}

	// pop without waiting, returning a null pointer if empty
	std::tr1::shared_ptr<AE_connection> pop_nowait()
	{
		pthread_mutex_lock(&mutex);
		std::tr1::shared_ptr<AE_connection> old =
		    queue.front();
		queue.pop();
		if (!queue.empty())
			pthread_cond_signal(&cond);
		pthread_mutex_unlock(&mutex);
		return old;
	}

	// pop with waiting
	std::tr1::shared_ptr<AE_connection> pop()
	{
		pthread_mutex_lock(&mutex);
		while (queue.empty())
			pthread_cond_wait(&cond, &mutex);
		std::tr1::shared_ptr<AE_connection> old =
		    queue.front();
		queue.pop();
		if (!queue.empty())
			pthread_cond_signal(&cond);
		pthread_mutex_unlock(&mutex);
		return old;
	}

private:
	// not copyable
	AE_queue(const AE_queue &) {}
	void operator=(const AE_queue &) {}

	std::queue<std::tr1::shared_ptr<AE_connection> > queue;
	pthread_cond_t	cond;
	pthread_mutex_t	mutex;
};

class AE_segment_queue {
public:
	typedef uint64_t				hash_type;

	AE_segment_queue(const struct sockaddr_in &addr)
	{
		std::vector<uint64_t> hashes;
		segment_hashes(addr, hashes);
		for (std::vector<uint64_t>::iterator i = hashes.begin();
		    i != hashes.end(); ++i) {
			_queue.push(*i);
		}
		pthread_mutex_init(&_lock, 0);
	}
	~AE_segment_queue()
	{
		pthread_mutex_destroy(&_lock);
	}

	void add_segment(const hash_type &hash)
	{
		lock();
		_queue.push(hash);
		unlock();
	}
	void remove_segment(const hash_type &hash)
	{
		// we won't need this
		assert(0);
	}

	hash_type acquire()
	{
		lock();
		hash_type hash = _queue.front();
		_queue.pop();
		_released.insert(hash);
		unlock();
		return hash;
	}
	void release(const hash_type &hash)
	{
		lock();
		std::set<hash_type>::iterator i = _released.find(hash);
		if (i != _released.end()) {
			_released.erase(i);
			_queue.push(hash);
		} else {
			assert(0);
		}
		unlock();
	}

private:
	// not copyable
	AE_segment_queue(const AE_segment_queue &) {}
	void operator=(const AE_segment_queue &) {}

	void lock()
	{
		pthread_mutex_lock(&_lock);
	}
	void unlock()
	{
		pthread_mutex_unlock(&_lock);
	}
	std::queue<hash_type>	_queue;
	std::set<hash_type>	_released;
	pthread_mutex_t		_lock;
};

#endif
