/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <cassert>
#include <cerrno>
#include <cstring>
#include <iostream>

#include "ae_queue.hpp"
#include "anti_entropy.hpp"

extern const char *prog;

namespace { // anon

struct ae_initial {
	char header[4]; // should be "ANT"
	uint8_t __padding[4];
	uint64_t seg_hash;
	uint64_t checksum;
};

// for full AE
struct ae_initial_response {
	// size of struct plus all entries
	uint32_t response_size;
	// count of entries, set to zero if nothing to do
	uint32_t entries;

	// next come entries, each:
	//	"key_string, null-terminated"
	//	data_version_struct
};

// for full AE
struct ae_request {
	// size of struct plus all entries
	uint32_t response_size;

	// count of request entries (what I want)
	uint32_t request_entries;

	// count of update entries (what you need)
	uint32_t update_entries;

	// next come request entries, each:
	//	"key_string, null-term"
	// next come update entries, each:
	//	"key_string, null-term"
	//	"value_string, null-term" .
	//	data_version_struct
};

struct ae_final {
	// size of struct plus all entries
	uint32_t response_size;

	// count of update entries (what you need)
	uint32_t update_entries;

	// next come update entries, each:
	//	"key_string, null-term"
	//	"value_string, null-term"
	//	data_version_struct
};


void		send_buf(int, const void *, size_t) throw (unix_error);
std::pair<uint8_t *, size_t>
		recv_buf(int) throw (server_error, unix_error);


void		send_init_resp(int, Segment *) throw (unix_error);
void		send_request(int, Segment *, const std::vector<std::string> &,
		    const std::vector<std::string> &) throw (unix_error);
void		send_final(int, Segment *, const std::vector<std::string> &)
		    throw (unix_error);


void		receive_init_resp(int, std::vector<std::string> *,
		    std::vector<data_version> &)
		    throw (server_error, unix_error);
void		receive_request(int, Segment *, std::vector<std::string> &)
		    throw (server_error, unix_error);
void		receive_final(int, Segment *)
		    throw (server_error, unix_error);


void
send_buf(int fd, const void *voidbuf, size_t size) throw (unix_error)
{
	const uint8_t *buf = reinterpret_cast<const uint8_t *>(voidbuf);
	while (size) {
		ssize_t bytes;
		if ((bytes = write(fd, buf, size)) == -1) {
			if (errno == EINTR) continue;
			throw unix_error("writing during anti-entropy");
		}
		buf += bytes;
		size -= bytes;
	}
}

std::pair<uint8_t *, size_t>
recv_buf(int fd) throw (server_error, unix_error)
{
	// first 32 bits is size
	uint32_t msg_size;
	uint8_t *sizebuf = reinterpret_cast<uint8_t *>(&msg_size);
	size_t size = 0;
	while (size < 4) {
		ssize_t bytes;
		if ((bytes = read(fd, sizebuf + size, 4 - size)) == -1) {
			if (errno == EINTR) continue;
			throw unix_error("reading during anti-entropy");
		} else if (!bytes)
			throw server_error(
			    "premature EOF during anti-entropy");
		sizebuf += bytes;
		size += bytes;
	}

	uint8_t *buf = new uint8_t[msg_size];
	std::copy(sizebuf, sizebuf + 4, buf);

	// receive remainder
	while (size < msg_size) {
		ssize_t bytes;
		if ((bytes = read(fd, buf + size, msg_size - size)) == -1) {
			if (errno == EINTR) continue;
			delete[] buf;
			throw unix_error("reading during anti-entropy");
		} else if (!bytes) {
			delete[] buf;
			throw server_error(
			    "premature EOF during anti-entropy");
		}
		sizebuf += bytes;
		size += bytes;
	}

	return std::pair<uint8_t *, size_t>(buf, msg_size);
}

void
send_init_resp(int fd, Segment *segment) throw (unix_error)
{
	typedef Segment::map_type map_type;
	map_type *map = segment ? segment->map() : 0;

	size_t size = sizeof(struct ae_initial_response);
	if (map) {
		for (map_type::iterator i = map->begin(); i != map->end();
		    ++i) {
			size += i->second->key().size() + 1;
			size += sizeof(struct data_version);
		}
	}

	// create buffer, write headers
	uint8_t *buf = new uint8_t[size];
	struct ae_initial_response *header =
	    reinterpret_cast<struct ae_initial_response *>(buf);
	header->response_size = size;
	header->entries = map ? map->size() : 0;

	if (map) {
		// append each entry
		uint8_t *pos = buf + sizeof(struct ae_initial_response);
		for (map_type::iterator i = map->begin(); i != map->end();
		    ++i) {
			const std::string &key = i->second->key();
			const data_version &vers = i->second->version();

			// append key
			const uint8_t *tmp =
			    reinterpret_cast<const uint8_t *>(key.c_str());
			std::copy(tmp, tmp + key.size() + 1, pos);
			pos += key.size() + 1;

			// append version
			tmp = reinterpret_cast<const uint8_t *>(&vers);
			std::copy(tmp, tmp + sizeof(data_version), pos);
			pos += sizeof(data_version);
		}
	}

	try {
		send_buf(fd, buf, size);
		delete[] buf;
	} catch (...) {
		delete[] buf;
		throw;
	}
}

void
send_request(int fd, Segment *segment,
    const std::vector<std::string> &requests,
    const std::vector<std::string> &updates) throw (unix_error)
{
	typedef Segment::map_type map_type;
	map_type *map = segment ? segment->map() : 0;

	size_t size = sizeof(struct ae_request);
	// what we need
	for (std::vector<std::string>::const_iterator i = requests.begin();
	    i != requests.end(); ++i) {
		size += i->size() + 1;
	}
	// what they need
	for (std::vector<std::string>::const_iterator i = updates.begin();
	    i != updates.end(); ++i) {
		const std::string &key = *i;
		const Data_item &item = *((*map)[key]);
		size += key.size() + 1;
		size += item.value().size() + 1;
		size += sizeof(struct data_version);
	}

	// create buffer, write headers
	uint8_t *buf = new uint8_t[size];
	struct ae_request *header = reinterpret_cast<struct ae_request *>(buf);
	header->response_size = size;
	header->request_entries = requests.size();
	header->update_entries = updates.size();

	// append each entry
	uint8_t *pos = buf + sizeof(struct ae_request);
	for (size_t i = 0; i < requests.size(); i++) {
		const std::string &key = requests[i];

		// append key
		const uint8_t *tmp =
		    reinterpret_cast<const uint8_t *>(key.c_str());
		std::copy(tmp, tmp + key.size() + 1, pos);
		pos += key.size() + 1;
	}
	for (size_t i = 0; i < updates.size(); i++) {
		const std::string &key = updates[i];
		const Data_item &item = *((*map)[key]);
		const std::string &value = item.value();
		const data_version &vers = item.version();

		// append key
		const uint8_t *tmp =
		    reinterpret_cast<const uint8_t *>(key.c_str());
		std::copy(tmp, tmp + key.size() + 1, pos);
		pos += key.size() + 1;

		// append value
		tmp = reinterpret_cast<const uint8_t *>(
		    value.c_str());
		std::copy(tmp, tmp + value.size() + 1, pos);
		pos += value.size() + 1;

		// append version
		tmp = reinterpret_cast<const uint8_t *>(&vers);
		std::copy(tmp, tmp + sizeof(data_version), pos);
		pos += sizeof(data_version);
	}

	try {
		send_buf(fd, buf, size);
		delete[] buf;
	} catch (...) {
		delete[] buf;
		throw;
	}
}

void
send_final(int fd, Segment *segment,
    const std::vector<std::string> &requested) throw (unix_error)
{
	typedef Segment::map_type map_type;
	map_type *map = segment ? segment->map() : 0;

	size_t size = sizeof(struct ae_final);
	for (std::vector<std::string>::const_iterator i = requested.begin();
	    i != requested.end(); ++i) {
		const std::string &key = *i;
		if(map->find(key) == map->end()){
			continue;
		}
		const Data_item &item = *((*map)[key]);
		const std::string &value = item.value();
		size += key.size() + 1;
		size += value.size() + 1;
		size += sizeof(struct data_version);
	}

	// create buffer, write headers
	uint8_t *buf = new uint8_t[size];
	struct ae_final *header = reinterpret_cast<struct ae_final *>(buf);
	header->response_size = size;
	header->update_entries = requested.size();

	// append each entry
	char *pos = reinterpret_cast<char *>(buf) + sizeof(struct ae_final);
	for (std::vector<std::string>::const_iterator i = requested.begin();
	    i != requested.end(); ++i) {
		const std::string &key = *i;
		if(map->find(key) == map->end()){
			continue;
		}
		const Data_item &item = *((*map)[key]);
		const std::string &value = item.value();
		const data_version &vers = item.version();

		// append key
		const uint8_t *tmp =
		    reinterpret_cast<const uint8_t *>(key.c_str());
		std::copy(tmp, tmp + key.size() + 1, pos);
		pos += key.size() + 1;

		// append value
		tmp = reinterpret_cast<const uint8_t *>(
		    value.c_str());
		std::copy(tmp, tmp + value.size() + 1, pos);
		pos += value.size() + 1;

		// append version
		tmp = reinterpret_cast<const uint8_t *>(&vers);
		std::copy(tmp, tmp + sizeof(data_version), pos);
		pos += sizeof(data_version);
	}

	try {
		send_buf(fd, buf, size);
		delete[] buf;
	} catch (...) {
		delete[] buf;
		throw;
	}
}

void
receive_init_resp(int fd, std::vector<std::string> &out_keys,
    std::vector<data_version> &out_versions) throw (server_error, unix_error)
{
	// receive entire message
	std::pair<uint8_t *, size_t> buf_pair = recv_buf(fd);
	uint8_t *buf = buf_pair.first;
	struct ae_initial_response *header =
	    reinterpret_cast<struct ae_initial_response *>(buf);
	size_t entries = header->entries;

	// for each entry:
	char *pos = reinterpret_cast<char *>(buf) +
	    sizeof(struct ae_initial_response);
	for (size_t i = 0; i < entries; i++) {
		out_keys.push_back(std::string(pos));
		pos += out_keys.back().size();

		data_version vers;
		std::copy(pos, pos + sizeof(vers),
		    reinterpret_cast<char *>(&vers));
		out_versions.push_back(vers);
		pos += sizeof(vers);
	}
	delete[] buf;
}

void
receive_request(int fd, Segment *segment,
    std::vector<std::string> &out_keys) throw (server_error, unix_error)
{
	using std::cerr;
	typedef Segment::map_type map_type;
	map_type *map = segment ? segment->map() : 0;

	// receive entire message
	std::pair<uint8_t *, size_t> buf_pair = recv_buf(fd);
	uint8_t *buf = buf_pair.first;
	struct ae_request *header =
	    reinterpret_cast<struct ae_request *>(buf);
	size_t update_entries = header->update_entries;
	size_t request_entries = header->request_entries;

	// for each request entry:
	char *pos = reinterpret_cast<char *>(buf) +
	    sizeof(struct ae_request);
	for (size_t i = 0; i < request_entries; i++) {
		out_keys.push_back(std::string(pos));
		pos += out_keys.back().size();
	}

	// for each update entry:
	for (size_t i = 0; i < update_entries; i++) {
		std::string key = pos;
		pos += key.size() + 1;

		std::string value = pos;
		pos += value.size() + 1;

		data_version vers;
		std::copy(pos, pos + sizeof(vers),
		    reinterpret_cast<char *>(&vers));
		pos += sizeof(vers);

		map_type::iterator iter = map->find(key);
#ifdef AE_DEBUG
		cerr << "AE-active: storing [" << key << "]=["  << value
		    << "]\n";
#endif
		if (iter == map->end()) {
			// add to map
			try {
				(*map)[key] = std::tr1::shared_ptr<Data_item>(
				    new Data_item(key, value, vers));
			} catch (...) {
				delete[] buf;
				throw;
			}
		} else
			// change value in map
			(*map)[key]->set(value, vers);
	}
	delete[] buf;
}

void
receive_final(int fd, Segment *segment) throw (server_error, unix_error)
{
	typedef Segment::map_type map_type;
	map_type *map = segment ? segment->map() : 0;

	// receive entire message
	std::pair<uint8_t *, size_t> buf_pair = recv_buf(fd);
	uint8_t *buf = buf_pair.first;
	struct ae_final *header = reinterpret_cast<struct ae_final *>(buf);
	size_t entries = header->update_entries;

	// for each entry:
	char *pos = reinterpret_cast<char *>(buf) + sizeof(struct ae_final);
	for (size_t i = 0; i < entries; i++) {
		std::string key = pos;
		pos += key.size() + 1;

		std::string value = pos;
		pos += value.size() + 1;

		data_version vers;
		std::copy(pos, pos + sizeof(vers),
		    reinterpret_cast<char *>(&vers));
		pos += sizeof(vers);

		map_type::iterator iter = map->find(key);
#ifdef AE_DEBUG
		cerr << "AE-active: storing [" << key << "]=[" << value
		    << "]\n";
#endif
		if (iter == map->end()) {
			// add to map
			try {
				(*map)[key] = std::tr1::shared_ptr<Data_item>(
				    new Data_item(key, value, vers));
			} catch (...) {
				delete[] buf;
				throw;
			}
		} else
			// change value in map
			(*map)[key]->set(value, vers);
	}
	delete[] buf;
}

} // end anon

void *
ae_respond(void *voidarg)
{
	using std::cerr;
	using std::ostringstream;
	using std::map;
	using std::string;
	using std::vector;

	ae_respond_args *args = reinterpret_cast<ae_respond_args *>(
	    voidarg);

	// Waits until somthing shows up in the ae queue, then sends all the
	// data related to the segment specified to the node that asked
	// for the data.

	// copy args
	AE_queue &queue = args->queue;
	Segment_map &seg_map = args->seg_map;

	typedef uint64_t				hash_type;
	typedef std::string				key_type;
	typedef std::tr1::shared_ptr<Data_item>		value_type;
	typedef std::map<key_type, value_type>		map_type;
	typedef std::tr1::shared_ptr<AE_connection>	ae_connect;

	for (;;) {
		//wait for request for data (a server just came back up)
		ae_connect connect1 = queue.pop();
		// closed automatically
		int sockfd = connect1->socket;

#ifdef AE_DEBUG
		cerr << "(AE-respond) receiving an ae_initial from primary\n";
#endif
		// just got a request (ANT, seg_hash they want, hash of
		// segment they want.)
		struct ae_initial initial;
		uint8_t *initial_buf = reinterpret_cast<uint8_t *>(&initial);
		try {
			// 3 bytes ("ANT") pulled off by listener
			size_t size = 3;
			while (size < sizeof(initial)) {
				ssize_t bytes;
				if ((bytes = read(sockfd, initial_buf + size,
				    sizeof(initial) - size)) == -1) {
					if (errno == EINTR) continue;
					throw unix_error(
					    "read of header failed");
				} else if (!bytes)
					throw server_error(
					    "premature EOF in header");
				size += bytes;
			}
		} catch (const general_error &e) {
			cerr << prog << "(AE-respond): " << e.what() << '\n';
			continue;
		}

#ifdef AE_DEBUG
		cerr << "(AE-respond) we think the seg_hash is: "
		    << initial.seg_hash << '\n';
#endif

		//get segment related to seg hash we were just asked for
		Segment *seg1 = 0;
		try {
			seg_map.link(initial.seg_hash);
			seg1 = seg_map.segment_acquire(initial.seg_hash);
		} catch (no_such_segment) {
		}
		assert(seg1);

		if (seg1->hash() == initial.checksum) {
#ifdef AE_DEBUG
			cerr << "(AE-respond) nothing to do\n";
#endif
			seg_map.segment_release(seg1);
			continue;
		}

#ifdef AE_DEBUG
		cerr << "(AE-respond) sending an ae_initial_response\n";
#endif
		try {
			send_init_resp(sockfd, seg1);
		} catch (const general_error &e) {
			cerr << prog << "(AE-respond): " << e.what() << '\n';
			if (seg1)
				seg_map.segment_release(seg1);
			continue;
		}

#ifdef AE_DEBUG
		cerr << "(AE-respond) receiving an ae_request\n";
#endif
		std::vector<std::string> requested_keys;
		try {
			receive_request(sockfd, seg1, requested_keys);
		} catch (const general_error &e) {
			cerr << prog << "(AE-respond): " << e.what() << '\n';
			if (seg1)
				seg_map.segment_release(seg1);
			continue;
		}

		try {
			send_final(sockfd, seg1, requested_keys);
		} catch (const general_error &e) {
			cerr << prog << "(AE-respond): " << e.what() << '\n';
			if (seg1)
				seg_map.segment_release(seg1);
			continue;
		}
		if (seg1)
			seg_map.segment_release(seg1); //clean up

#ifdef AE_DEBUG
		cerr << "(AE-respond) done\n";
#endif
	}

	// abort since no looping
	return 0;
}

void *
ae_active(void *voidarg)
{
	using std::cerr;
	using std::ostringstream;
	using std::string;
	using std::vector;
	typedef Segment::map_type map_type;

	ae_active_args *args = reinterpret_cast<ae_active_args *>(
	    voidarg);

	// copy args
	AE_segment_queue &seg_queue = args->seg_queue;
	Segment_map &seg_map = args->seg_map;
	Segment_owner_map &owner_map = args->segment_owner_map;
	typedef uint64_t	hash_type;

	for (;;) {
		hash_type seg_hash = seg_queue.acquire();

#ifdef AE_DEBUG
		cerr << "(AE-active) we think the seg_hash is "
		    << seg_hash << "\n";
#endif
		//get a ptr to a segment related to the seg hash
		Segment* seg1 = seg_map.segment_acquire(seg_hash); //Jeff's first line of code
		map_type *map = seg1->map();

		//find node with your data on it Segement_owner_map.successors(seg_hash)
		//get back a vector with addresses
		vector<struct sockaddr_in *> successors = owner_map.successors(seg_hash);
		if (successors.size() < 2) {
			// just give up, since there will never be more added
			seg_map.segment_release(seg1);
			seg_queue.release(seg_hash);
			return 0;
		}

		//get address of secondary node (use recently obtained vector of node addresses)
		
		//I'm guessing you want the second element of the array?
		struct sockaddr_in *node_with_data = successors[1];
		//offset size of sock struct to get secondary

		//compose message to send to secondary node (ANT, seg_hash, hash of segment)
		struct ae_initial request;
		// zero out the padding
		memset(request.__padding, 0, sizeof(request.__padding));
		memcpy(request.header, "ANT", 4);
		request.seg_hash = seg_hash;
		request.checksum = seg1->hash();
		char *out = reinterpret_cast<char *>(&request);

		// keys/versions secondary has
		std::vector<std::string> other_keys;
		std::vector<data_version> other_versions;
		std::set<std::string> other_key_set;
		// what we want
		std::vector<std::string> wanted;
		// what they need
		std::vector<std::string> tosend;

#ifdef AE_DEBUG
		cerr << "(AE-active) sending ae_initial to secondary\n";
#endif
		//actually send the message to the node with data
		int sock = -1;
		try {
			sock = send_to_addr_n(out, sizeof(request),
			    node_with_data);
		} catch (const unix_error &e) {
			if (e.errnum() != ECONNREFUSED)
				cerr << prog << "(AE-active): " << e.what()
				    << "\n(this is fine provided the server "
				    << "is down)\n";
			goto end_of_loop;
		}

#ifdef AE_DEBUG
		cerr << "(AE-active) receiving ae_initial_response from "
		    "secondary\n";
#endif
		try {
			receive_init_resp(sock, other_keys, other_versions);
		} catch (const unix_error &e) {
			cerr << prog << "(AE-active): " << e.what() << '\n';
			goto end_of_loop;
		} catch (const server_error &e) {
#ifdef AE_DEBUG
			cerr << prog << "(AE-active): " << e.what() << '\n';
			cerr << "(probably the segments are the same)\n";
#endif
			goto end_of_loop;
		}

		for (size_t i = 0; i < other_keys.size(); i++) {
			map_type::iterator j = map->find(other_keys[i]);
			if (j == map->end()) {
				// secondary has, I don't
				wanted.push_back(other_keys[i]);
			} else if (other_versions[i] > j->second->version()) {
				// secondary has newer
				wanted.push_back(other_keys[i]);
			} else if (other_versions[i] < j->second->version()) {
				// secondary has older
				tosend.push_back(other_keys[i]);
			}
			other_key_set.insert(other_keys[i]);
		}
		// find keys the secondary doesn't have
		for (map_type::iterator i = map->begin(); i != map->end();
		    ++i) {
			if (other_key_set.find(i->first) ==
			    other_key_set.end()) {
				tosend.push_back(i->first);
			}
		}

#ifdef AE_DEBUG
		cerr << "(AE-active) sending ae_request to secondary\n";
#endif
		try {
			send_request(sock, seg1, wanted, tosend);
		} catch (const general_error &e) {
			cerr << prog << "(AE-active): " << e.what() << '\n';
			goto end_of_loop;
		}

#ifdef AE_DEBUG
		cerr << "(AE-active) receiving ae_final from secondary\n";
#endif
		try {
			receive_final(sock, seg1);
		} catch (const general_error &e) {
			cerr << prog << "(AE-active): " << e.what() << '\n';
			goto end_of_loop;
		}
#ifdef AE_DEBUG
		cerr << "(AE-active) done\n";
#endif

end_of_loop:
		if (sock != -1)
			close(sock);

		// drop locks
		seg_map.segment_release(seg1);

		// put segment back on the queue
		seg_queue.release(seg_hash);

		// sleep until next request
		nanosleep(&SLEEP_AFTER_AE, 0);
	}

	return 0;
}
