/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

// unix
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

// c++
#include <algorithm>
#include <cassert>
#include <cerrno>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <limits>
#include <memory>
#include <vector>
#include <stdint.h>

// libc
#include <getopt.h>

#include "client_comm.hpp"
#include "parse.hpp"

const char *prog;

namespace {

void
usage()
{
	std::cout <<
"usage: " << prog << " -k KEY [-v VALUE | -V VALUE_FILE]\n"
"OPTIONS\n"
"  --help              Print this usage information\n"
"  -h HOSTS            Ignored\n"
"  -k KEY              Key to read/write\n"
"  -v VALUE            Optional; value to write\n"
"  -V VALUE_FILE       Optional; value file to write\n";
}

std::pair<uint8_t *, size_t>
read_whole_file(const std::string &path) throw (client_error, unix_error)
{
	// this is kind of janky since I cannot use boost
	struct stat	st;
	size_t		remaining;
	uint8_t		*value_pos;
	uint8_t		*value;

	if (stat(path.c_str(), &st) == -1)
		throw unix_error("failed to stat value file");

	value = new uint8_t[st.st_size];

	std::ifstream file(path.c_str(), std::ios_base::in |
	    std::ios_base::binary);
	if (!file) {
		// sort of assuming that errno is set
		delete[] value;
		throw unix_error("open failed for value file");
	}

	remaining = st.st_size;
	value_pos = value;
	while (remaining) {
		size_t bytes =
		    remaining < static_cast<size_t>(st.st_blksize) ?
		    remaining : st.st_blksize;
		file.read(reinterpret_cast<char *>(value_pos), bytes);
		if (file.eof()) {
			delete[] value;
			throw client_error("premature EOF for value file");
		} else if (file.fail()) {
			delete[] value;
			throw unix_error("read failed for value file");
		}
		value_pos += file.gcount();
		remaining -= file.gcount();
	}

	// make sure no bytes are left
	file.get();
	if (!file.eof()) {
		delete[] value;
		throw client_error("file increased in size while reading");
	}

	return std::make_pair(value, st.st_size);
}

std::pair<std::string, uint16_t>
split_hostname(const std::string &host) throw (parse_error)
{
	using std::pair;
	using std::string;

	string::const_iterator hostname_end = std::find(
	    host.begin(), host.end(), ':');
	if (hostname_end == host.end())
		throw parse_error("hostname parsing failed");

	string::const_iterator port_begin = hostname_end + 1;

	string port_str(port_begin, host.end());
	char *endptr;
	errno = 0;
	long port_num = strtol(port_str.c_str(), &endptr, 10);
	if (errno || *endptr || port_num <= 0 ||
	    port_num > std::numeric_limits<uint16_t>::max())
		throw parse_error("port number invalid");

	return pair<string, uint16_t>(
	    string(host.begin(), hostname_end), port_num);
}

}

int
main(int argc, char **argv)
{
	using std::auto_ptr;
	using std::cout;
	using std::cerr;
	using std::pair;
	using std::string;
	using std::vector;

	vector<pair<string, uint16_t> > hosts;
	auto_ptr<pair<string, uint16_t> > explicit_host;
	auto_ptr<string> key;
	auto_ptr<string> value_str;
	auto_ptr<string> value_path;

	prog = *argv;

	if (argc == 1) {
		usage();
		return 0;
	}

	// we'll print errors ourselves
	opterr = 0;

	// use large values otherwise not returned by getopt_long()
	enum { OPT_HELP = 256 };
	struct option longopts[] = {
		{ "help", 0, 0, OPT_HELP },
		{ 0, 0, 0, 0 }
	};

	for (;;) {
		// leading '-' makes order of positional options not matter;
		// next ':' causes missing option arguments to return ':'
		int opt = getopt_long(argc, argv, "-:h:k:v:V:", longopts, 0);
		if (opt == -1) break;

		switch (opt) {
		case OPT_HELP:
			usage();
			return 0;
		case 'h':
			// ignore the host:port
		/*
			// push each argument onto vector until the next
			// option
			hosts.push_back(optarg);
			while (optind < argc && argv[optind][0] != '-')
				hosts.push_back(argv[optind++]);
		*/
			break;
		case 'k':
			key.reset(new string(optarg));
			break;
		/*
		case 'p': {
			long	l;
			char	*endptr;
			errno = 0;
			l = strtol(optarg, &endptr, 10);
			if (errno || *endptr) {
				cerr << prog
				    << ": error parsing port number\n";
				return 1;
			}
			// [1,65535]
			if (l <= 0 ||
			    l > std::numeric_limits<uint16_t>::max()) {
				cerr << prog << ": bad port number\n";
				return 1;
			}
			port = l;
			break;
		}
		*/
		case 'v': // value string
			value_path.reset();
			value_str.reset(new string(optarg));
			break;
		case 'V': // value file
			value_str.reset();
			value_path.reset(new string(optarg));
			break;
		case 1: // positional
			if (explicit_host.get()) {
				cerr << prog <<
				    ": too many explicit hosts specified\n";
				return 1;
			}
			try {
				// no, this is not used ever
				explicit_host.reset(
				    new pair<string, uint16_t>(
				    split_hostname(optarg)));
			} catch (const parse_error &e) {
				cerr << prog << ": " << e.what() << '\n';
				return 1;
			}
			break;
		case ':':
			cerr << prog << ": missing argument for -"
			    << static_cast<char>(optopt) << '\n';
			return 1;
		case '?':
			cerr << prog << ": unrecognized option -"
			    << static_cast<char>(optopt) << '\n';
			return 1;
		default:
			assert(0);
		}
	}
	if (!key.get()) {
		cerr << prog << ": no key specified\n";
		return 1;
	}

	/*
	try {
		config_parse(CLIENTD_CONFIG, hosts);
	} catch (const general_error &e) {
		cerr << prog << ": " << e.what() << '\n';
		return 1;
	}

	if (hosts.empty()) {
		cerr << prog << ": no hosts specified\n";
		return 1;
	}
	*/

	pair<uint8_t *, size_t> value_pair(0, 0);
	if (value_path.get())
		try {
			value_pair = read_whole_file(*value_path);
		} catch (const general_error &e) {
			cerr << prog << ": " << e.what() << '\n';
			return 1;
		}
	else if (value_str.get()) {
		value_pair.first = new uint8_t[value_str->size()];
		value_pair.second = value_str->size();
		std::copy(value_str->begin(), value_str->end(),
		    value_pair.first);
	}

	pair<uint8_t *, size_t> result;
	try {
		result = make_request_clientd(hosts, *key,
		    value_pair.first, value_pair.second);
		delete[] value_pair.first;
	} catch (const key_undefined &) {
		// a mild error
		delete[] value_pair.first;
		cout << "key is undefined\n";
		return 1;
	} catch (const general_error &e) {
		// bad error
		delete[] value_pair.first;
		cerr << prog << ": " << e.what() << '\n';
		return 1;
	}

	cout << '[' << *key << "]=";
	if (result.second) {
		cout << '[';
		for (size_t i = 0; i < result.second; i++)
			cout.put(result.first[i]);
		cout << "]\n";
		delete[] result.first;
	} else {
		cout << "NIL\n";
	}

	return 0;
}
