/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include <cassert>
#include <cerrno>
#include <cstring>
#include <iostream>
#include <limits>
#include <sstream>

#include "client_comm.hpp"
#include "parse.hpp"

extern const char *prog;

// anonymous namespace
namespace {

const std::string LISTEN_SOCKET = "/tmp/clientd.socket";
const size_t CLIENT_BACKLOG = 100;

// just close a file descriptor when out of scope
struct fildes_owner {
	fildes_owner(int fd) : fd(fd) {}

	// explicitly close
	void close()
	{
		::close(fd);
		fd = -1;
	}

	~fildes_owner()
	{
		if (fd != -1) close();
	}

	int fd;

private:
	// non-copyable
	fildes_owner(const fildes_owner &) {}
	void operator=(const fildes_owner &) {}
};

template <typename T>
struct array_owner {
	array_owner(T *ptr) : ptr(ptr) {}

	// explicitly delete
	void dealloc()
	{
		delete[] ptr;
		ptr = 0;
	}
	~array_owner()
	{
		if (ptr) dealloc();
	}
	T *ptr;

private:
	// non-copyable
	array_owner(const array_owner &) {}
	void operator=(const array_owner &) {}
};

struct clientd_request {
	// total length
	uint32_t	size;
	uint8_t		have_value;
	uint8_t		num_hosts;
	uint8_t		__padding[2];
	uint32_t	value_size;
	// next come port numbers (uint16_t[]), padded for 0 mod 2 elements
	// next come num_hosts hostnames, all null-terminated
	// next comes the key, null-terminated
	// next comes the value if present
};

struct clientd_response {
	// total length
	uint32_t	size;
	uint32_t	value_size;
	uint16_t	http_status; // 0 if no connection made
	uint8_t		status; // enum response_status
	uint8_t		have_value;
	// next comes the value if present
};

// allocates/returns a buffer of the message
std::pair<uint8_t *, size_t>
read_message(int fd) throw (bad_message, unix_error)
{
	uint32_t size;
	uint8_t *size_buf = reinterpret_cast<uint8_t *>(&size);

	// first 4 bytes always the size
	ssize_t bytes;
	size_t pos = 0;
	while (pos < 4) {
		bytes = read(fd, size_buf + pos, 4 - pos);
		if (!bytes)
			throw bad_message("premature end-of-file");
		else if (bytes == -1) {
			if (errno == EINTR) continue;
			throw unix_error("reading message");
		}
		pos += bytes;
	}

	uint8_t *buf = new uint8_t[size];
	std::copy(size_buf, size_buf + 4, buf);

	// read remainder
	try {
		while (pos < size) {
			bytes = read(fd, buf + pos, size - pos);
			if (!bytes)
				throw bad_message("premature end-of-file");
			else if (bytes == -1) {
				if (errno == EINTR) continue;
				throw unix_error("reading message");
			}
			pos += bytes;
		}
	} catch (...) {
		delete[] buf;
		throw;
	}

	return std::make_pair(buf, size);
}

}

std::pair<uint8_t *, size_t>
make_request_clientd(
    const std::vector<std::pair<std::string, uint16_t> > &hosts,
    const std::string &key, const uint8_t *value, size_t value_size)
    throw (bad_message, client_error, key_undefined, server_error, unix_error)
{
	using std::pair;
	using std::string;
	using std::vector;

	size_t request_size = sizeof(struct clientd_request) + key.size() +
	    1 + value_size;

	for (vector<pair<string, uint16_t> >::const_iterator i =
	    hosts.begin(); i != hosts.end(); ++i)
		// 1 null byte, two port bytes
		request_size += i->first.size() + 3;
	// padding
	if (hosts.size() % 2) request_size += 2;

	uint8_t request_buf[request_size];
	struct clientd_request *request =
	    reinterpret_cast<struct clientd_request *>(request_buf);
	request->have_value = value ? 1 : 0;
	request->value_size = value ? value_size : 0;
	request->size = request_size;
	request->num_hosts = hosts.size();

	// copy ports
	uint8_t *pos = request_buf + sizeof(struct clientd_request);
	for (vector<pair<string, uint16_t> >::const_iterator i =
	    hosts.begin(); i != hosts.end(); ++i) {
		*reinterpret_cast<uint16_t *>(pos) = i->second;
		pos += 2;
	}
	if (hosts.size() % 2) pos += 2;

	// copy hosts
	for (vector<pair<string, uint16_t> >::const_iterator i =
	    hosts.begin(); i != hosts.end(); ++i) {
		std::copy(i->first.begin(), i->first.end(),
		    reinterpret_cast<char *>(pos));
		pos[i->first.size()] = 0;
		pos += i->first.size() + 1;
	}

	// copy key + null
	std::copy(key.c_str(), key.c_str() + key.size() + 1,
	    reinterpret_cast<char *>(pos));
	pos += key.size() + 1;

	// copy value
	if (value) {
		std::copy(value, value + value_size,
		    reinterpret_cast<char *>(pos));
		pos += value_size;
	}

	int sock;
	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
		throw unix_error("opening socket to client daemon");
	fildes_owner sock_owner(sock);

	// connect
	{
		struct sockaddr_un addr;
		addr.sun_family = AF_UNIX;
		std::copy(LISTEN_SOCKET.begin(), LISTEN_SOCKET.end() + 1,
		    addr.sun_path);

		if (connect(sock, reinterpret_cast<struct sockaddr *>(&addr),
		    sizeof(addr)) == -1) {
			if (errno == ENOENT)
				throw client_error(
				    "client daemon not running");
			throw unix_error("connecting to client daemon");
		}
	}

	// write request
	{
		ssize_t bytes;
		size_t written = 0;
		while (written < request_size) {
			bytes = write(sock, request_buf + written,
			    request_size - written);
			if (bytes == -1) {
				if (errno == EINTR) continue;
				unix_error("writing to client daemon socket");
			}
			written += bytes;
		}
	}

	// retrieve result
	std::pair<uint8_t *, size_t> response_buf = read_message(sock);
	array_owner<uint8_t> result_owner(response_buf.first);

	struct clientd_response *response =
	    reinterpret_cast<struct clientd_response *>(response_buf.first);
	assert(response_buf.second >= sizeof(struct clientd_response));
	assert(response_buf.second == response->size);

	switch (response->status) {
	case STATUS_SUCCESS:
		break;
	case STATUS_KEY_UNDEFINED:
		throw key_undefined();
	case STATUS_ERROR: {
		std::ostringstream out;
		out << "Server returned " << response->http_status;
		throw server_error(out.str());
	}
	case STATUS_CLIENTD_ERROR:
		throw client_error();
	case STATUS_CONN_ERROR:
		throw server_error("Connection error");
	case STATUS_UNAVAILABLE:
		throw server_error("Service unavailable");
	default:
		assert(0);
	}

	if (!response->have_value)
		return std::make_pair(static_cast<uint8_t *>(0), 0);
	else {
		uint8_t *copy;
		if (response->value_size) {
			copy = new uint8_t[response->value_size];
			uint8_t *start = response_buf.first +
			    sizeof(struct clientd_response);
			std::copy(start, start + response->value_size, copy);
		} else
			copy = 0;
		return std::make_pair(copy, response->value_size);
	}
}

// clientd: start listening, return socket
int
start_listening() throw (unix_error)
{
	int sockfd;

	if ((sockfd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
		throw unix_error("creating listen socket");

	struct sockaddr_un addr;
	addr.sun_family = AF_UNIX;
	std::copy(LISTEN_SOCKET.begin(), LISTEN_SOCKET.end() + 1,
	    addr.sun_path);

	// remove the file if it exists
	unlink(LISTEN_SOCKET.c_str());

	if (bind(sockfd, reinterpret_cast<struct sockaddr *>(&addr),
	    sizeof(addr)) == -1) {
		long error = errno;
		close(sockfd);
		throw unix_error("binding unix socket", error);
	}

	if (listen(sockfd, CLIENT_BACKLOG) == -1) {
		long error = errno;
		close(sockfd);
		throw unix_error("listen failed", error);
	}

	return sockfd;
}

void *
clientd_listener(void *voidargs) throw (client_error, unix_error)
{
	using std::cerr;
	using std::pair;
	using std::string;
	using std::vector;
	using std::tr1::shared_ptr;

	// copy arguments to local references
	clientd_listener_args *args =
	    reinterpret_cast<clientd_listener_args *>(voidargs);
	Clientd_queue	&queue = args->queue;
	int		sockfd = args->socket;

	for (;;) {
		int fd;
		struct sockaddr_un addr;
		socklen_t addrlen = sizeof(addr);

		if ((fd = accept(sockfd,
		    reinterpret_cast<struct sockaddr *>(&addr), &addrlen))
		    == -1) {
			cerr << prog << "(listener): accept failed: "
			    << strerror(errno) << '\n';
			continue;
		}

		// read size
		uint32_t	size;
		uint8_t		*size8 = reinterpret_cast<uint8_t *>(&size);
		size_t		count = 0;
		while (count < 4) {
			ssize_t bytes;
			if ((bytes = read(fd, size8 + count, 4 - count))
			    == -1) {
				if (errno == EINTR) continue;
				cerr << prog << "(listener): read failed: "
				    << strerror(errno) << '\n';
				close(fd);
				fd = -1;
				break;
			} else if (!bytes) {
				cerr << prog << "(listener): premature EOF\n";
				close(fd);
				fd = -1;
				break;
			}
			count += bytes;
		}
		if (fd == -1) continue;

		// read rest of message
		uint8_t buf[size];
		*reinterpret_cast<uint32_t *>(buf) = size;
		while (count < size) {
			ssize_t bytes;
			if ((bytes = read(fd, buf + count, size - count))
			    == -1) {
				if (errno == EINTR) continue;
				cerr << prog << "(listener): read failed: "
				    << strerror(errno) << '\n';
				close(fd);
				fd = -1;
				break;
			} else if (!bytes) {
				cerr << prog << "(listener): premature EOF\n";
				close(fd);
				fd = -1;
				break;
			}
			count += bytes;
		}
		if (fd == -1) continue;

		struct clientd_request *request =
		    reinterpret_cast<struct clientd_request *>(buf);

		// get ports
		uint8_t *pos = buf + sizeof(struct clientd_request);
		uint16_t *ports = reinterpret_cast<uint16_t *>(pos);
		pos += 2 * request->num_hosts;
		if (request->num_hosts % 2) pos += 2;

		// get hosts
		vector<pair<string, uint16_t> > hostnames;
		for (size_t i = 0; i < request->num_hosts; i++) {
			hostnames.push_back(std::make_pair(
			    reinterpret_cast<char *>(pos),
			    ports[i]));
			pos += hostnames.back().first.size() + 1;
		}

		// resolve hosts
		vector<struct sockaddr_in> hosts;
		resolve_hosts(hostnames, hosts);

		// get key
		string key = reinterpret_cast<char *>(pos);
		pos += key.size() + 1;

		// get value
		uint8_t *value;
		if (request->have_value) {
			try {
				value = new uint8_t[request->value_size];
			} catch (...) {
				close(fd);
				throw;
			}
			std::copy(pos, pos + request->value_size, value);
		} else
			value = 0;

//#ifdef CLIENT_DEBUG
		if (value) {
			cerr << "listener: received request put(" << key
			    << ',';
			for (size_t i = 0; i < request->value_size; i++)
				cerr.put(value[i]);
			cerr << ")\n";
		} else
			cerr << "listener: received request get(" << key
			    << ")\n";
//#endif

		// done! add to queue
		queue.push(shared_ptr<Clientd_connection>(
		    new Clientd_connection(hosts, fd, key,
		    value, request->value_size)));
	}

	// thread shouldn't exit
	return 0;
}

void
send_response(const uint8_t *value, size_t value_size,
    enum response_status status, uint16_t http_status,
    const Clientd_connection &cx)
    throw (unix_error)
{
	size_t total = sizeof(struct clientd_response) + value_size;
	uint8_t buf[total];
	struct clientd_response *response =
	    reinterpret_cast<clientd_response *>(buf);
	response->size = total;
	response->value_size = value_size;
	response->status = status;
	response->http_status = http_status;
	response->have_value = !!value;
	std::copy(value, value + value_size,
	    buf + sizeof(struct clientd_response));

	size_t count = 0;
	while (count < total) {
		ssize_t bytes;
		if ((bytes = write(cx.socket, buf + count, total - count))
		    == -1)
			throw unix_error("write failed");
		count += bytes;
	}
}
