/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <sys/types.h>
#include <sys/wait.h>
#include <poll.h>
#include <signal.h>
#include <unistd.h>

#include <cassert>
#include <cerrno>
#include <cstdlib>
#include <iostream>
#include <memory>
#include <vector>

#include <getopt.h>

#include "client_comm.hpp"
#include "clientd_view.hpp"
#include "clientd_worker.hpp"
#include "errors.hpp"
#include "parse.hpp"

const char *prog;

namespace {

const unsigned NUM_WORKERS = 20;

struct thread_owner {
	thread_owner(pthread_t thread=0) : thread(thread) {}
	~thread_owner()
	{
		if (thread) pthread_kill(thread, SIGKILL);
	}

	pthread_t release()
	{
		pthread_t old = thread;
		thread = 0;
		return old;
	}

	pthread_t thread;

private:
	// non-copyable
	thread_owner(const thread_owner &) {}
	void operator=(const thread_owner &) {}
};

void
usage()
{
	std::cout << "usage: " << prog << " [-d]\n"
"OPTIONS\n"
"  -d          Daemonize\n"
"  -h, --help  Show this message\n";
}

void
clientd_setup(const std::vector<struct sockaddr_in> &addrs,
    int completion_fd=-1) throw (unix_error)
{
	using std::vector;
	using std::tr1::shared_ptr;

	// shared data structures
	Clientd_queue		queue;
	Clientd_host_view	view(addrs.begin(), addrs.end());
	Segment_owner_map	owner_map(addrs);

	int	listen_socket;
	int	errval;

	listen_socket = start_listening();

	// create listener
	pthread_t listener;
	clientd_listener_args clientd_listener_args(queue, listen_socket);
	if ((errval = pthread_create(&listener, 0,
	    clientd_listener, &clientd_listener_args)))
		throw unix_error("creating listener thread", errval);
	thread_owner listener_owner(listener);

	// create workers
	vector<pthread_t> workers;
	vector<shared_ptr<clientd_worker_args> > worker_args;
	vector<shared_ptr<thread_owner> > worker_owners;
	for (unsigned i = 0; i < NUM_WORKERS; i++) {
		pthread_t worker;
		shared_ptr<clientd_worker_args> args(new clientd_worker_args(
		    queue, view, owner_map, i));

		if ((errval = pthread_create(&worker, 0,
		    clientd_worker, args.get())))
			throw unix_error("creating worker thread", errval);
		shared_ptr<thread_owner> owner(new thread_owner(worker));

		workers.push_back(worker);
		worker_args.push_back(args);
		worker_owners.push_back(owner);
	}

	// create probing thread
	pthread_t probe;
	clientd_probe_args probe_args(view);
	if ((errval = pthread_create(&probe, 0, clientd_probe, &probe_args)))
		throw unix_error("creating probe thread", errval);
	thread_owner probe_owner(probe);

	// done with setup
	if (completion_fd != -1) {
		char c = 0;
		if (write(completion_fd, &c, 1) == -1)
			throw unix_error("failed to write to pipe");
	}

	std::cerr << ">>> clientd listening for requests\n";

	// effectively, never exit
	pthread_join(listener, 0);
	listener_owner.release();
	for (size_t i = 0; i < workers.size(); i++) {
		pthread_join(workers[i], 0);
		worker_owners[i]->release();
	}
	pthread_join(probe, 0);
	probe_owner.release();

	// should never get here
	std::cerr << "something awful happened\n";
	exit(1);
}

pid_t
start_child(const std::vector<struct sockaddr_in> &addrs) throw (unix_error)
{
	// TODO this function does not do a very good job deallocating
	// resources on error; I'm thinking of using same method as in
	// client_comm.cpp, but I hate it

	int	status_pipe[2];
	pid_t	child;

	if (pipe(status_pipe) == -1)
		throw unix_error("creating pipe");

	if ((child = fork()) == -1) {
		throw unix_error("forking child");
	} else if (!child) {
		close(status_pipe[0]);
		clientd_setup(addrs, status_pipe[1]);
		assert(0);
	}
	close(status_pipe[1]);

	// loop until child either exits or completes initialization
	struct pollfd fd;
	fd.fd = status_pipe[0];
	fd.events = POLLIN;
	for (;;) {
		int	retval;
		int	status;

		// check if child is done initializing
		retval = poll(&fd, 1, 0);
		if (retval == -1) {
			if (errno != EINTR) {
				long err = errno;
				kill(child, SIGTERM);
				throw unix_error(
				    "error polling child process", err);
			}
		} else if (retval) {
			char byte;
			if (read(status_pipe[0], &byte, 1) == -1) {
				throw unix_error("error reading pipe");
			}
			// child is now done
			break;
		}

		// see if child has exited/terminated
		retval = waitpid(child, &status, WNOHANG);
		if (retval == -1) {
			if (errno != EINTR) {
				long err = errno;
				kill(child, SIGTERM);
				throw unix_error("error waiting for child",
				    err);
			}
		} else if (retval == child) {
			if (WIFEXITED(status) &&
			    WEXITSTATUS(status)) {
				throw client_error(
				    "child exited with non-zero status");
			} else if (WIFEXITED(status)) {
				throw client_error(
				    "child exited successfully???");
			} else if (WIFSIGNALED(status)) {
				throw client_error(
				    "child was terminated with a signal");
			}
			assert(0);
		}

		// let child finish
		sched_yield();
	}

	return child;
}

}

int
main(int argc, char **argv)
{
	using std::auto_ptr;
	using std::cerr;
	using std::cout;
	using std::pair;
	using std::vector;
	using std::string;

	bool	daemonize = false;

	prog = *argv;

	// we'll print errors ourselves
	opterr = 0;

	// use large values otherwise not returned by getopt_long()
	enum { OPT_HELP = 256 };
	struct option longopts[] = {
		{ "help", 0, 0, OPT_HELP },
		{ 0, 0, 0, 0 }
	};

	for (;;) {
		// leading ':' causes missing option arguments to return ':'
		int opt = getopt_long(argc, argv, ":dh", longopts, 0);
		if (opt == -1) break;

		switch (opt) {
		case 'd':
			daemonize = true;
			break;
		case OPT_HELP:
		case 'h':
			usage();
			return 0;
		case ':':
			cerr << prog << ": missing argument for -"
			    << static_cast<char>(optopt) << '\n';
			return 1;
		case '?':
			cerr << prog << ": unrecognized option -"
			    << static_cast<char>(optopt) << '\n';
			return 1;
		default:
			assert(0);
		}
	}

	vector<struct sockaddr_in> addrs;
	try {
		vector<pair<string, uint16_t> > hosts;
		config_parse(CLIENTD_CONFIG, hosts);
		resolve_hosts(hosts, addrs);
	} catch (const general_error &e) {
		cerr << prog << ": " << e.what() << '\n';
		return 1;
	}

	if (addrs.empty()) {
		cerr << prog << ": no hosts specified\n";
		return 1;
	}

	if (daemonize) {
		pid_t pid = start_child(addrs);
		cout << "Daemonized to pid " << pid << '\n';
	} else
		clientd_setup(addrs);

	return 0;
}
