/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef CLIENTD_QUEUE_HPP
#define CLIENTD_QUEUE_HPP

#include <netinet/in.h>
#include <pthread.h>
#include <unistd.h>

#include <queue>
#include <string>
#include <tr1/memory>
#include <stdint.h>

struct Clientd_connection {
	Clientd_connection(
	    const std::vector<struct sockaddr_in> &hosts,
	    int socket, const std::string &key,
	    const uint8_t *value, size_t value_size) :
		hosts(hosts),
		key(key),
		value_size(value_size),
		socket(socket)
	{
		if (value) {
			this->value = new uint8_t[value_size];
			std::copy(value, value + value_size, this->value);
		} else
			this->value = 0;
	}
	~Clientd_connection()
	{
		delete[] value;
		close(socket);
	}
	std::vector<struct sockaddr_in> hosts;
	std::string	key;
	size_t		value_size;
	uint8_t		*value;
	int		socket;

private:
	// non-copyable
	Clientd_connection(const Clientd_connection &) {}
	void operator=(const Clientd_connection &) {}
};

// a synchronized queue
class Clientd_queue {
	static const size_t MAX_SIZE = 500;
public:
	Clientd_queue()
	{
		pthread_cond_init(&cond_nonempty, 0);
		pthread_cond_init(&cond_nonfull, 0);
		pthread_mutex_init(&mutex, 0);
	}
	~Clientd_queue()
	{
		pthread_cond_destroy(&cond_nonempty);
		pthread_cond_destroy(&cond_nonfull);
		pthread_mutex_destroy(&mutex);
	}

	void push(std::tr1::shared_ptr<Clientd_connection> cx)
	{
		pthread_mutex_lock(&mutex);
		while (queue.size() == 500)
			pthread_cond_wait(&cond_nonfull, &mutex);

		queue.push(cx);

		pthread_cond_signal(&cond_nonempty);
		if (queue.size() != 500)
			pthread_cond_signal(&cond_nonfull);

		pthread_mutex_unlock(&mutex);
	}

	// pop without waiting, returning a null pointer if empty
	std::tr1::shared_ptr<Clientd_connection> pop_nowait()
	{
		pthread_mutex_lock(&mutex);

		std::tr1::shared_ptr<Clientd_connection> old =
		    queue.front();
		queue.pop();

		pthread_cond_signal(&cond_nonfull);
		if (!queue.empty())
			pthread_cond_signal(&cond_nonempty);

		pthread_mutex_unlock(&mutex);
		return old;
	}

	// pop with waiting
	std::tr1::shared_ptr<Clientd_connection> pop()
	{
		pthread_mutex_lock(&mutex);
		while (queue.empty())
			pthread_cond_wait(&cond_nonempty, &mutex);

		std::tr1::shared_ptr<Clientd_connection> old =
		    queue.front();
		queue.pop();

		pthread_cond_signal(&cond_nonfull);
		if (!queue.empty())
			pthread_cond_signal(&cond_nonempty);

		pthread_mutex_unlock(&mutex);
		return old;
	}

private:
	// not copyable
	Clientd_queue(const Clientd_queue &) {}
	void operator=(const Clientd_queue &) {}

	std::queue<std::tr1::shared_ptr<Clientd_connection> > queue;
	pthread_cond_t	cond_nonempty;
	pthread_cond_t	cond_nonfull;
	pthread_mutex_t	mutex;
};

#endif
