/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef CLIENTD_VIEW_HPP
#define CLIENTD_VIEW_HPP

#include <netinet/in.h>
#include <pthread.h>

#include <cstdlib>
#include <map>
#include <string>
#include <vector>
#include <stdint.h>
#include <tr1/memory>

#include "clock.h"
#include "segment.hpp"

//const uint64_t VIEW_STILL_UP = 1000000ULL * ;

struct Sockaddrin_cmp {
	bool operator()(const struct sockaddr_in &a,
	    const struct sockaddr_in &b) const
	{
		const uint8_t *a8 = reinterpret_cast<const uint8_t *>(&a);
		const uint8_t *b8 = reinterpret_cast<const uint8_t *>(&b);
		for (size_t i = 0; i < sizeof(struct sockaddr_in); i++)
			if (a8[i] != b8[i])
				return a8[i] < b8[i];
		return false;
	}
};

class Clientd_host {
public:
	Clientd_host(const struct sockaddr_in &addr,
	    uint64_t timestamp) :
		_addr(addr),
		_last_up(timestamp),
		_last_down(timestamp)
	{}

	void mark(bool up)
	{
		uint64_t now = clock_monotonic();
		if (up)
			_last_up = now;
		else
			_last_down = now;
	}

	const struct sockaddr_in &addr() const
	{
		return _addr;
	}
	bool is_up() const
	{
		return _last_up >= _last_down;
	}
	uint64_t downtime(uint64_t now=0) const
	{
		if (_last_up > _last_down) return 0;
		if (!now) now = clock_monotonic();
		return now - _last_down;
	}

private:
	// not copyable
	Clientd_host(const Clientd_host &) {}
	void operator=(const Clientd_host &) {}

	struct sockaddr_in	_addr;
	uint64_t		_last_up;
	uint64_t		_last_down;
};

class Clientd_host_view {
public:
	typedef uint64_t				hash_type;
	typedef struct sockaddr_in			key_type;
	typedef std::tr1::shared_ptr<Clientd_host>	value_type;
	typedef std::map<key_type, value_type, Sockaddrin_cmp>
							map_type;

	template <typename In>
	Clientd_host_view(In begin, In end)
	{
		uint64_t now = clock_monotonic();
		for (In i = begin; i != end; ++i)
			_map[*i] = std::tr1::shared_ptr<Clientd_host>(
			    new Clientd_host(*i, now));

		pthread_rwlock_init(&_mutex, 0);
	}
	Clientd_host_view()
	{
		pthread_rwlock_init(&_mutex, 0);
	}
	~Clientd_host_view() {}

	// add() ignores entries that already exist
	template <typename In>
	void add(In begin, In end)
	{
		wrlock();
		uint64_t now = clock_monotonic();
		while (begin != end)
			private_add(*begin++, now);
		unlock();
	}
	void add(const struct sockaddr_in &host)
	{
		wrlock();
		private_add(host, clock_monotonic());
		unlock();
	}

	// choose the first 'up' node in the list; if no nodes are 'up', then
	// the one with the longest known downtime is chosen (if it turns out
	// to be down, mark() it down and try again)
	const struct sockaddr_in *select(
	    const std::vector<struct sockaddr_in *> &hosts) const
	{
		assert(hosts.size());

		const struct sockaddr_in *addr = 0;
		const struct sockaddr_in *last_resort = 0;
		uint64_t last_resort_downtime = 0;

		rdlock();
		uint64_t now = clock_monotonic();

		for (std::vector<struct sockaddr_in *>::const_iterator i =
		    hosts.begin(); i != hosts.end(); ++i) {
			map_type::const_iterator j = _map.find(**i);
			assert(j != _map.end());
			if (j->second->is_up()) {
				addr = *i;
				break;
			}
			uint64_t downtime = j->second->downtime(now);
			if (downtime > last_resort_downtime) {
				last_resort = *i;
				last_resort_downtime = downtime;
			}
		}
		unlock();

		assert(addr || last_resort);
		return addr ? addr : last_resort;
	}

	void mark(const key_type &key, bool up)
	{
		rdlock();
		map_type::iterator i = _map.find(key);
		unlock();

		assert(i != _map.end());

		// pure write, no synchronization
		i->second->mark(up);
	}

	size_t count() const
	{
		return _map.size();
	}

private:
	void rdlock() const
	{
		Clientd_host_view *this_ = const_cast<Clientd_host_view *>(
		    this);
		pthread_rwlock_rdlock(&this_->_mutex);
	}
	void wrlock()
	{
		pthread_rwlock_wrlock(&_mutex);
	}
	void unlock() const
	{
		Clientd_host_view *this_ = const_cast<Clientd_host_view *>(
		    this);
		pthread_rwlock_unlock(&this_->_mutex);
	}
	void private_add(const struct sockaddr_in &addr, uint64_t timestamp)
	{
		value_type host(new Clientd_host(addr, timestamp));
		std::vector<hash_type> hashes;
		segment_hashes(addr, hashes);
		for (std::vector<hash_type>::iterator i = hashes.begin();
		    i != hashes.end(); ++i)
			_map.insert(std::make_pair(addr, host));
	}

	// non-copyable
	Clientd_host_view(const Clientd_host_view &) {}
	void operator=(const Clientd_host_view &) {}

	map_type		_map;

	pthread_rwlock_t	_mutex;
};

struct clientd_probe_args {
	clientd_probe_args(Clientd_host_view &view) : view(view) {}

	Clientd_host_view	&view;
};

void	*clientd_probe(void *voidargs);

#endif
