/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <unistd.h>

#include <cerrno>
#include <climits>
#include <cstring>
#include <iostream>
#include <sstream>

#include "client_comm.hpp"
#include "clientd_worker.hpp"
#include "clock.h"
#include "errors.hpp"
#include "http.hpp"
#include "util.hpp"

extern const char *prog;

namespace {

std::string
hostname() throw (unix_error)
{
	char buf[HOST_NAME_MAX + 1];
	if (gethostname(buf, HOST_NAME_MAX) == -1)
		throw unix_error("gethostname() failed");
	buf[HOST_NAME_MAX] = '\0';
	return std::string(buf);
}

int
web_connect(const struct sockaddr_in *host) throw (unix_error)
{
	using std::pair;
	using std::string;
	using std::vector;

	int sock;

#ifdef CLIENT_DEBUG
	std::cerr << "connecting to " << addr_to_str(*host) << '\n';
#endif

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		throw unix_error("creating socket");

	if (connect(sock, reinterpret_cast<const struct sockaddr *>(host),
	    sizeof(struct sockaddr_in)) == -1) {
		long error = errno;
		close(sock);
		std::ostringstream out;
		out << "connecting to " << addr_to_str(*host);
		throw unix_error(out.str(), error);
	}

	return sock;
}

std::string
thread_name(unsigned number)
{
	std::ostringstream out;
	out << prog << "(worker " << number << ')';
	return out.str();
}

void
short_response(const std::string &name, Clientd_connection &cx,
    enum response_status status)
{
	using std::cerr;

	try {
		send_response(0, 0, status, 0, cx);
	} catch (const unix_error &e) {
		cerr << name << ": " << e.what() << '\n';
	}
}

// read everything from a socket until EOF
std::string
receive_response(int sock) throw (unix_error)
{
	using std::ostringstream;

	const size_t BUFSIZE = 1024;
	ostringstream out;
	char buf[BUFSIZE + 1];
	for (;;) {
		ssize_t bytes;
		if ((bytes = read(sock, buf, BUFSIZE)) == -1) {
			if (errno == EINTR) continue;
			throw unix_error("reading from server socket");
		} else if (!bytes)
			break;

		buf[bytes] = '\0';
		out << buf;
	}

	return out.str();
}

// has the socket been disconnected?
bool
still_alive(int sock) throw (unix_error)
{
	// DONTWAIT: do not block
	// PEEK: don't take bytes off the buffer
	ssize_t bytes;
	char byte;
	if ((bytes = recv(sock, &byte, 1, MSG_DONTWAIT | MSG_PEEK)) == -1) {
		switch (errno) {
		case EAGAIN: return true;

		// this doesn't seem to happen
		case ENOTCONN: return false;
		}
		throw unix_error("recv failed to detect state");
	}
	return bytes;
}

std::tr1::shared_ptr<Http_response>
perform_transaction(const std::string &name, Clientd_connection &cx,
    Clientd_host_view &view, Segment_owner_map &owner_map,
    const std::string &message)
    throw (http_parse_error, unix_error)
{
	using std::cerr;
	using std::string;
	using std::vector;
	using std::tr1::shared_ptr;

	Segment_owner_map::hash_type seg_hash =
	    owner_map.find_segment(cx.key);
	vector<struct sockaddr_in *> addrs = owner_map.successors(seg_hash);

#ifdef CLIENT_DEBUG
	cerr << "responsible addrs:";
	for (size_t i = 0; i <addrs.size(); i++)
		cerr << ' ' << addr_to_str(*addrs[i]);
	cerr << '\n';
#endif

	uint8_t attempts = 0;
	// create connection
	for (;;) {
		if (!still_alive(cx.socket))
			throw unix_error("client connection lost", ENOTCONN);

		if (attempts++ == view.count()) {
			attempts = 0;
			nanosleep(&SLEEP_ALLFAIL, 0);
			continue;
		}

		int websock;
		const sockaddr_in *addr = view.select(addrs);
		try {
			websock = web_connect(addr);
		} catch (const unix_error &e) {
			cerr << name << ": " << e.what() << '\n';
			view.mark(*addr, false);
			continue;
		}

		// send request to server
		size_t written = 0;
		while (written < message.size()) {
			ssize_t bytes;
			if ((bytes = write(websock, message.c_str() + written,
			    message.size() - written)) == -1) {
				if (errno == EINTR) continue;
				cerr << name << ": socket write error"
				    << strerror(errno) << '\n';
				view.mark(*addr, false);
				break;
			}
			written += bytes;
		}
		if (written < message.size()) {
			close(websock);
			view.mark(*addr, false);
			continue;
		}

		string response_str;
		try {
			response_str = receive_response(websock);
		} catch (const unix_error &e) {
			cerr << name << ": " << e.what() << '\n';
			close(websock);
			view.mark(*addr, false);
			continue;
		}

		shared_ptr<Http_response> response;
		try {
			response.reset(new Http_response("", response_str));
		} catch (const http_parse_error &e) {
			cerr << name << ": " << e.what() << '\n';
			close(websock);
			view.mark(*addr, false);
			continue;
		}

		close(websock);
		return response;
	}
}

} // end anon namespace

void *
clientd_worker(void *voidargs)
{
	using std::cerr;
	using std::ostringstream;
	using std::string;
	using std::vector;
	using std::tr1::shared_ptr;

	clientd_worker_args *args = reinterpret_cast<clientd_worker_args *>(
	    voidargs);

	// copy args
	Clientd_host_view	&view = args->view;
	Clientd_queue		&queue = args->queue;
	Segment_owner_map	&owner_map = args->owner_map;

	string		name = thread_name(args->worker_number);

	for (;;) {
		shared_ptr<Clientd_connection> cx = queue.pop();

		string self_name;
		try {
			self_name = hostname();
		} catch (const unix_error &e) {
			cerr << name << ": " << e.what() << '\n';
			short_response(name, *cx, STATUS_CLIENTD_ERROR);
			continue;
		}

		ostringstream out;
		out << "GET [" << cx->key << ']';

		// XXX though not necessary according to requrements,
		// keys and values could be %-escaped
		if (cx->value) {
			out << "=[";
			for (size_t i = 0; i < cx->value_size; i++)
				out.put(cx->value[i]);
			out << ']';
		}

		out << " HTTP/1.1\r\n"
		    << "Host: " << self_name << "\r\n"
		    << "User-Agent: " << USER_AGENT << "\r\n"
		    << "Connection: close\r\n\r\n";

		string message = out.str();

		shared_ptr<Http_response> response;
		try {
			response = perform_transaction(name, *cx, view,
			    owner_map, message);
		} catch (const general_error &e) {
			cerr << name << ": " << e.what() << '\n';
			short_response(name, *cx, STATUS_CONN_ERROR);
			continue;
		}

		// mark servers up if the server indicated so (this should
		// include the server we just talked to)
		const vector<struct sockaddr_in> &addrs =
		    response->available_servers();
		for (vector<struct sockaddr_in>::const_iterator i =
		    addrs.begin(); i != addrs.end(); ++i)
			view.mark(*i, true);

		const uint8_t *value = response->have_value() ?
		    reinterpret_cast<const uint8_t *>(
		    response->value().c_str()) : 0;
		size_t value_size = response->have_value() ?
		    response->value().size() : 0;

		if (response->status() >= 500 && response->status() < 600) {
			short_response(name, *cx, STATUS_UNAVAILABLE);
			continue;
		}

		if (!response->have_value() && !cx->value) {
			short_response(name, *cx, STATUS_KEY_UNDEFINED);
			continue;
		}

		try {
			// indicate success or vague error
			enum response_status status =
			    response->status() == 200 ?
			    STATUS_SUCCESS : STATUS_ERROR;

			send_response(value, value_size, status,
			    response->status(), *cx);
#ifdef CLIENT_DEBUG
			cerr << "worker " << args->worker_number
			    << ": responded to "
			    << (cx->value ? "put(" : "get(")
			    << cx->key << ')';
			if (value)
				cerr << ", got `" << value << '\'';
			cerr << '\n';
#endif
		} catch (const unix_error &e) {
			cerr << name << ": " << e.what() << '\n';
		}
	}

	return 0;
}
