/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef ERRORS_HPP
#define ERRORS_HPP

#include <exception>
#include <string>
#include <stdint.h>

// superclass for all our exceptions
struct general_error : std::exception {};

struct bad_message : general_error {
	// implicitly use errno
	bad_message(const std::string &msg);

	virtual ~bad_message() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct client_error : general_error {
	client_error(const std::string &msg="");

	virtual ~client_error() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct http_parse_error : general_error {
	http_parse_error(const std::string &msg,
	    unsigned line=static_cast<unsigned>(-1));

	virtual ~http_parse_error() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct key_undefined : general_error {
	key_undefined() {}

	virtual ~key_undefined() throw () {}
	virtual const char *what() const throw ()
	{
		return "key undefined";
	}
};

struct lookup_error : general_error {
	// errnum should be h_errno (see gethostbyname(3))
	lookup_error(const std::pair<std::string, uint16_t> &host,
	    int errnum);

	virtual ~lookup_error() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct need_lock : std::exception {
	need_lock() {}
	virtual ~need_lock() throw () {}
	virtual const char *what() const throw ()
	{
		return "need lock";
	}
};

struct no_such_segment : general_error {
	no_such_segment() : _msg("no such segment") {}

	virtual ~no_such_segment() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct parse_error : general_error {
	parse_error(const std::string &msg, unsigned line=0);

	virtual ~parse_error() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct request_restart : general_error {
	request_restart(const std::string &msg);

	virtual ~request_restart() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct server_error : general_error {
	server_error(const std::string &msg="");

	virtual ~server_error() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}

	std::string _msg;
};

struct unix_error : general_error {
	// implicitly use errno
	unix_error(const std::string &msg);
	unix_error(const std::string &msg, int errnum);

	virtual ~unix_error() throw () {}
	virtual const char *what() const throw ()
	{
		return _msg.c_str();
	}
	int errnum() const
	{	return _errnum; }

private:
	std::string	_msg;
	int		_errnum;
};

#endif
