/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <sys/select.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <cstring>
#include <cerrno>

#include "heartbeat.hpp"
#include "errors.hpp"

void *
heartbeat_monitor(void *voidarg)
{
	heartbeat_monitor_args *args =
	    reinterpret_cast<heartbeat_monitor_args *>(voidarg);

	// copy args
	Server_table &servers = args->servers;

	struct sockaddr_in me, sender;
	int sockin;
	struct heartbeat hb;
	int recvMsgSize;
	unsigned int senderLen; //used in recv calls
	uint16_t inport = servers.get_self().sin_port;

	//here we set up a nice udp socket for receiving
	// Setup local socket address
	memset(&me, 0, sizeof(me));
	me.sin_family = AF_INET;
	me.sin_port = inport;
	me.sin_addr.s_addr = htonl(INADDR_ANY);
	
	//create the socket
	if ((sockin = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
		assert(0);
	
	int rebind = 1;
	if (setsockopt(sockin, SOL_SOCKET, SO_REUSEADDR, &rebind,
	    sizeof(rebind)) == -1) {
		long error = errno;
		close(sockin);
		throw unix_error("reusing the address", error);
	}

	//Bind the socket for incoming data
	if (bind(sockin, (struct sockaddr *) &me, sizeof(me)) < 0){
		//bind failed
		throw unix_error("heartbeat bind failure");
		assert(0);
	}
	
	// now we endlessly loop, updating the server table every time we see
	// a new packet
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(sockin, &rfds);

	//check if there's a message, and get it if so
	while (select(sockin+1, &rfds, 0, 0, 0) != -1) {
		senderLen = sizeof(sender);
		if ((recvMsgSize = recvfrom(sockin, &hb, sizeof(hb), 0, 
			 (struct sockaddr *) &sender, &senderLen)) < 0){
			//recv failed
			assert(0);
		}
		//set the port to the "correct" value in the sender sockaddr
		sender.sin_port = hb.port; //let's pretend everything is in the right
		 //endianness, because honestly, it is.
		servers.update_timestamp(sender, hb.timestamp);

		// reset rfds, just in case
		FD_SET(sockin, &rfds);
	}
	assert(0); //help something failed
}

void *
heartbeat_maker(void *voidarg)
{
	heartbeat_maker_args *args =
	    reinterpret_cast<heartbeat_maker_args *>(voidarg);

	// copy args
	Server_table &servers = args->servers;

	// get self
	struct sockaddr_in self_addr = servers.get_self();

	//we need one for each server.  Maybe I'll make a vector.
	std::vector<std::pair<int, sockaddr_in> > out_sockets;
	std::vector<const struct sockaddr_in *> nodes = servers.get_nodes();
	std::vector<const struct sockaddr_in *>::const_iterator it;
	for (it=nodes.begin(); it < nodes.end(); it++){
		//check if this is me
		if (!servers.compare_sockaddr_in(*(*it), self_addr)){
			std::pair<int, sockaddr_in> tmp;
			out_sockets.push_back(tmp);
			out_sockets.back().second = *(*it);
			//set up the sockets
			if ((out_sockets.back().first = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
				assert(0);
		}
	}
	
	// Then we loop through them!  We don't want to use too much cpu so
	// we sleep a lot
	while (true){
		std::vector<std::pair<int, sockaddr_in> >:: iterator sock_it;
		nanosleep(&HEARTBEAT_INTERVAL, 0);
		struct heartbeat hb;
		// memset to keep valgrind from complaining
		memset(&hb, 0, sizeof(hb));
		hb.timestamp = clock_monotonic();
		hb.port = servers.get_self().sin_port;
		// iterate through vector
		for(sock_it=out_sockets.begin(); sock_it < out_sockets.end(); sock_it++){
			//compare current time with last response
			if (hb.timestamp - servers.get_last_timestamp(
			    sock_it->second) > HEARTBEAT_TOLERANCE) {
				// mark as down
				servers.update_up(sock_it->second, false);
			}
			//send the current time
			if (sendto(sock_it->first, &hb, sizeof(struct heartbeat), 0,
				(struct sockaddr *) &(sock_it->second), sizeof(sock_it->second)) < 0){
					assert(0);
			}
		}
	}
	return 0;
}
