/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <cassert>
#include <cerrno>
#include <cstring>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include <pcrecpp.h>

#include "http.hpp"
#include "parse.hpp"
#include "segment.hpp"

Http_response::Http_response(const std::string &pfx, int sock)
    throw (http_parse_error, unix_error)
{
	using std::ostringstream;

	ostringstream out;
	out << pfx;
	size_t size = pfx.size();
	for (;;) {
		char	c;
		ssize_t bytes;

		if ((bytes = read(sock, &c, 1)) == -1) {
			if (errno == EINTR) continue;
			throw unix_error("read failed");
		} else if (!bytes)
			throw http_parse_error("premature EOF");
		size++;
		out << c;

		// end of headers; lower limit actually much larger than 4,
		// but whatevs
		if (size > 4 && out.str().substr(size - 4, size)
		    == "\r\n\r\n")
			break;
	}
	parse(out.str());

	out.flush();
	size = 0;
	while (size < _size) {
		size_t SIZE = 4096;
		char buf[SIZE+1];

		size_t readsize = (_size - size) < 4096 ?
		    (_size - size) : 4096;

		ssize_t bytes;
		if ((bytes = read(sock, buf, readsize)) == -1) {
			if (errno == EINTR) continue;
			throw unix_error("read failed");
		} else if (!bytes)
			throw http_parse_error("premature EOF");
		size += bytes;
		buf[bytes] = '\0';
		out << buf;
	}

	parse_content(out.str());
}

void
Http_response::parse(const std::string &str) throw (http_parse_error)
{
	using namespace pcrecpp;
	using std::istringstream;
	using std::pair;
	using std::vector;

	RE re_status("^ HTTP/(.+) \\s+ (\\d+) \\s+ (.+)",
	    RE_Options().set_extended(true));

	RE re_field("^ (.+?) \\s* : \\s* (.+)",
	    RE_Options().set_extended(true));

	RE re_version("^ (\\d+) , (\\d+) , (\\d+) , (\\d+) , (\\d+)",
	    RE_Options().set_extended(true));

	istringstream in(str);
	string line;

	if (!std::getline(in, line))
		throw http_parse_error("unexpected end of message", 0);

	string version;
	unsigned status;
	string status_name;
	if (!re_status.PartialMatch(line, &version, &status, &status_name))
		throw http_parse_error("bad status line", 1);

	// we don't care about the version/status_name now, but we do
	// the status
	_response = status;

	// http prop-value pairs (vector to maintain order)
	vector<pair<string, string> > properties;

	_linenum = 1;
	_size = 0;
	for (;;) {
		if (!std::getline(in, line))
			throw http_parse_error("unexpected end of message",
			    _linenum);
		_linenum++;
		if (line == "\r")
			break;

		string property;
		string value;
		if (!re_field.PartialMatch(line, &property, &value))
			throw http_parse_error("bad property", _linenum);

		if (value[value.size()-1] == '\r')
			value = value.substr(0, value.size() - 1);

		properties.push_back(std::make_pair(property, value));

		if (!strcasecmp(property.c_str(), "Content-length")) {
			std::istringstream in(value);
			in >> _size;

		} else if (!strcasecmp(property.c_str(), "Pragma") &&
		    !strncasecmp(value.c_str(), "servers=", 8)) {
			string remainder = value.substr(8, value.size());
			vector<pair<string, uint16_t> > available;
			pragma_servers_parse(remainder, available);
			resolve_hosts(available, _available_servers);

		} else if (!strcasecmp(property.c_str(), "Pragma") &&
		    !strncasecmp(value.c_str(), "version=", 8)) {
			string remainder = value.substr(8, value.size());

			std::tr1::shared_ptr<data_version> vers(
			    new data_version());
			
			if (!re_version.PartialMatch(remainder,
			    vers->versions+0, vers->versions+1,
			    vers->versions+2, vers->versions+3,
				&vers->timestamp)){
#ifdef HTTP_DEBUG
			    	std::cerr << "Bad version, remainder was "
				    << remainder << "\n";
#endif
				throw http_parse_error(
				    "bad version Http_response::parse",
				    _linenum);
			}
			_version = vers;
		}

	}
}

void
Http_response::parse_content(const std::string &str) throw (http_parse_error)
{
	using namespace pcrecpp;
	using std::istringstream;
	using std::pair;
	using std::vector;

	RE re_keyval("^ \\[ (.+) \\] = (?: (NIL) | \\[ (.*) \\] )",
	    RE_Options().set_extended(true));

	istringstream in(str);

	bool have_key = false;
	_have_value = false;
	for (;;) {
		std::string line;
		if (!std::getline(in, line)) break;
		_linenum++;

		if (!have_key) {
			string nil;
			if (re_keyval.PartialMatch(line, &_key, &nil,
			    &_value)) {
				if (nil.empty())
					_have_value = true;
				have_key = true;
			}
		}

		size_t end = line.find('\r');
		_content += line.substr(0, end) + '\n';
	}
}

const data_version &
Http_response::version() const
{
	return *_version;
}

void
Http_request::parse(const std::string &str) throw (http_parse_error)
{
	using namespace pcrecpp;
	using std::istringstream;
	using std::pair;
	using std::vector;

	RE re_status(
	    "^ GET \\s+ \\[ (.+?) \\] (?:(=) \\[ (.+) \\] )? \\s+ "
	    "HTTP / (.+)",
	    RE_Options().set_extended(true));

	RE re_field("^ (.+) : \\s* (.+)",
	    RE_Options().set_extended(true));

	RE re_version("^ (\\d+) , (\\d+) , (\\d+) , (\\d+) , (\\d+)",
	    RE_Options().set_extended(true));

	istringstream in(str);
	string line;

	if (!std::getline(in, line))
		throw http_parse_error("unexpected end of message", 0);

	string equal;
	string version;
#ifdef DEBUG_HTTP
	std::cerr << "The request is " << str << "\n";
#endif
	if (!re_status.PartialMatch(line, &_key, &equal, &_value, &version))
		throw http_parse_error("bad request", 1);
	_have_value = !equal.empty();

	// http prop-value pairs (vector to maintain order)
	vector<pair<string, string> > properties;

	_no_forwarding = false;
	unsigned linenum = 1;
	for (;;) {
		if (!std::getline(in, line))
			throw http_parse_error("unexpected end of message",
			    linenum);
		linenum++;
		if (line == "\r")
			break;

		string property;
		string value;
		if (!re_field.PartialMatch(line, &property, &value))
			throw http_parse_error("bad property", linenum);

		if (value[value.size()-1] == '\r')
			value = value.substr(0, value.size() - 1);

		properties.push_back(std::make_pair(property, value));
		if (!strcasecmp(property.c_str(), "User-Agent"))
			_client = value;

		if (!strcasecmp(property.c_str(), "Pragma") &&
		    !strncasecmp(value.c_str(), "no_forwarding=", 14)) {
			string remainder = value.substr(14, value.size());
			if (!strcasecmp(remainder.c_str(), "true"))
				_no_forwarding = true;
		} else if (!strcasecmp(property.c_str(), "Pragma") &&
		    !strncasecmp(value.c_str(), "version=", 8)) {
			string remainder = value.substr(8, value.size());

			std::tr1::shared_ptr<data_version> vers(
			    new data_version());
			
			if (!re_version.PartialMatch(remainder,
			    vers->versions+0, vers->versions+1,
			    vers->versions+2, vers->versions+3,
				&vers->timestamp)) {
#ifdef HTTP_DEBUG
				std::cout << "Bad version, remainder was "
				    << remainder << "\n";
#endif
				throw http_parse_error(
				    "bad version Http_request::parse",
				    linenum);
			}
			
			_version = vers;
		}
	}
}

Http_request::Http_request(const std::string &pfx, int sock)
    throw (http_parse_error, unix_error)
{
	using std::ostringstream;

	ostringstream out;
	out << pfx;
	size_t size = pfx.size();
	for (;;) {
		char	c;
		ssize_t bytes;

		if ((bytes = read(sock, &c, 1)) == -1) {
			if (errno == EINTR) continue;
			throw unix_error("read failed");
		} else if (!bytes)
			throw http_parse_error("premature EOF");
		size++;
		out << c;

		// end of headers; lower limit actually much larger than 4,
		// but whatevs
		if (size > 4 && out.str().substr(size - 4, size)
		    == "\r\n\r\n")
			break;
	}

	parse(out.str());
}

const data_version &
Http_request::version() const
{
	return *_version;
}
