/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef HTTP_HPP
#define HTTP_HPP

#include <netinet/in.h>

#include <stdint.h>
#include <string>
#include <vector>
#include <tr1/memory>

#include "errors.hpp"

// The Peloquin-Roller-Olson (PRO) client
const char *const USER_AGENT = "PRO client";
const char *const SERVER_USER_AGENT = "PRO server";

class data_version;

// parses an HTTP response
class Http_response {
public:
	Http_response(const std::string &pfx, const std::string &str)
	    throw (http_parse_error)
	{
		std::string msg = pfx + str;
		size_t end = msg.find("\r\n\r\n");
		end += 4;
		parse(msg.substr(0, end));
		parse_content(msg.substr(end, msg.size()));
	}
	Http_response(const std::string &pfx, int fd)
	    throw (http_parse_error, unix_error);

	unsigned status() const
	{
		return _response;
	}
	// you probably want key(), have_value(), value() below
	// should be ${KEY}=${VALUE} or ${KEY}=NIL
	const std::string &content() const
	{
		return _content;
	}
	const std::vector<struct sockaddr_in> &available_servers() const
	{
		return _available_servers;
	}

	const std::string &key() const
	{
		return _key;
	}
	bool have_value() const
	{
		return _have_value;
	}
	const std::string &value() const
	{
		return _value;
	}
	const data_version &version() const;

private:
	void parse(const std::string &str) throw (http_parse_error);
	void parse_content(const std::string &str) throw (http_parse_error);

	std::string _content;
	unsigned _response;
	std::vector<struct sockaddr_in> _available_servers;
	std::string _key;
	std::string _value;
	size_t	_size;
	unsigned _linenum;
	bool _have_value;
	std::tr1::shared_ptr<data_version> _version;
};

// parses an HTTP request
class Http_request {
public:
	Http_request(const std::string &pfx, const std::string &str)
	    throw (http_parse_error)
	{
		parse(pfx.size() ? pfx+str : str);
	}
	Http_request(const std::string &pfx, int sock)
	    throw (http_parse_error, unix_error);

	// compare against USER_AGENT and SERVER_USER_AGENT (otherwise, a
	// client and not ours)
	const std::string &client() const
	{
		return _client;
	}

	const std::string &key() const
	{
		return _key;
	}
	bool have_value() const
	{
		return _have_value;
	}
	const std::string &value() const
	{
		return _value;
	}
	const bool &no_forwarding() const
	{
		return _no_forwarding;
	}
	const data_version &version() const;

private:
	void parse(const std::string &str) throw (http_parse_error);

	std::string _client;
	std::string _key;
	std::string _value;
	bool _have_value;
	bool _no_forwarding;
	std::tr1::shared_ptr<data_version> _version;
};

#endif
