/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <netdb.h>

#include <cassert>
#include <cerrno>
#include <cstring>
#include <fstream>
#include <iostream>

#include <pcrecpp.h>

#include "parse.hpp"

/******************************************************************************
* resolve_hosts
* takes:    &hosts
*           &out_addrs
* does:		sets up the sockaddr_in
******************************************************************************/
void
resolve_hosts(const std::vector<std::pair<std::string, uint16_t> > &hosts,
    std::vector<struct sockaddr_in> &out_addrs)
    throw (lookup_error)
{
	const size_t	INCR = 64;
	static size_t	bufsz = INCR;
	char		*buf;

	buf = new char[bufsz];

	for (size_t i = 0; i < hosts.size(); i++) {
		struct sockaddr_in	addr;
		struct hostent		ent;
		struct hostent		*hp;
		int			herr;

		while (gethostbyname_r(hosts[i].first.c_str(),
		    &ent, buf, bufsz,
		    &hp, &herr) == ERANGE) {
			// increase size of buffer and try again
			delete[] buf;
			bufsz += INCR;
			buf = new char[bufsz];
		}
		if (!hp) {
			delete[] buf;
			throw lookup_error(hosts[i], herr);
		}

		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = htons(hosts[i].second);
		char *addr_ptr = reinterpret_cast<char *>(&addr.sin_addr);
		std::copy(hp->h_addr_list[0], hp->h_addr_list[0] + 4,
		    addr_ptr);

		out_addrs.push_back(addr);
	}
	delete[] buf;
}

/******************************************************************************
* configParse
* takes:    &path  The path to the file listing hosts
*           &out_hosts  A vector of pairs of hostnames and ports
* does:     parses the file, updates out_hosts
******************************************************************************/
void
config_parse(const std::string &path,
	std::vector<std::pair<std::string, uint16_t> > &out_hosts)
	throw (parse_error, unix_error)
{
	pcrecpp::RE line_exp("^ ([^:]+) : (\\d+) $",
	    pcrecpp::RE_Options().set_extended(true));

	std::ifstream inFile(path.c_str());
	if (inFile.fail()) {
		throw unix_error("open failed for config file");
	}
	unsigned lineno = 0;
	for (;;) {
		std::string line;
		if (!std::getline(inFile, line)) break;
		lineno++;
		if (line == "\r") continue;

		std::string hostname;
		uint16_t port;
		if (!line_exp.FullMatch(line, &hostname, &port))
			throw parse_error("bad line", lineno);
		out_hosts.push_back(std::make_pair(hostname, port));
	}
	inFile.close();
}

void
pragma_servers_parse(const std::string &value,
    std::vector<std::pair<std::string, uint16_t> > &out_hosts)
    throw (parse_error)
{
	using namespace pcrecpp;
	using std::pair;
	using std::string;
	using std::vector;

	// remove a possible prefix "Pragma: servers=X", "servers=X", or "X",
	RE pfx_regex("^ (?:(?: Pragma: \\s* )? servers= )? ",
	    RE_Options().set_extended(true));

	RE hostnames(
	    "      ([^:,]+) : (\\d+) [,$] ",
	    //"(?: , ([^:,]+) : (\\d+) )*",
	    RE_Options().set_extended(true));

	StringPiece piece = value;

	pfx_regex.Consume(&piece);

	for (;;) {
		string hostname;
		uint16_t port;
		int consumed;

		// what an awful interface; why cannot the FullMatch()
		// function just take a vector of Arg?
		Arg arg0 = &hostname;
		Arg arg1 = &port;
		Arg *args[] = { &arg0, &arg1 };

		if (!hostnames.DoMatch(piece, RE::ANCHOR_START, &consumed,
		    args, 2)) break;
		piece.remove_prefix(consumed);

		out_hosts.push_back(pair<string, uint16_t>(hostname, port));
	}
}
