/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <iostream>
#include <set>

#include "md5.hpp"
#include "segment.hpp"

struct Sockaddrin_cmp {
	bool operator()(const sockaddr_in &a, const sockaddr_in &b) const
	{
		if (a.sin_port != b.sin_port)
			return a.sin_port < b.sin_port;
		if (a.sin_addr.s_addr != b.sin_addr.s_addr)
			return a.sin_addr.s_addr < b.sin_addr.s_addr;
		return false;
	}
};


void
segment_hashes(const struct sockaddr_in &addr, std::vector<uint64_t> &hashes)
{
	for (uint8_t i = 0; i < SEGMENTS_PER_NODE; i++) {
		MD5_hash md5;
		md5.update(&i, 1);
		md5.update(&(addr.sin_port), sizeof(addr.sin_port));
		md5.update(&(addr.sin_addr.s_addr), sizeof(addr.sin_addr.s_addr));
		//md5.update(&addr, sizeof(struct sockaddr_in));
		hashes.push_back(md5.final64());
#ifdef SEGMENT_DEBUG
		std::cerr << "we just stored " << hashes.back() << "\n";
#endif
	}
}

void
data_version::cmp(const data_version &rhs, bool &greater, bool &less) const
{
	less = true;
	greater = true;
	for (size_t i = 0; i < NUM_HOSTS; i++) {
		if (versions[i] <= rhs.versions[i]) {
			greater = false;
			if (!less) break;
		}
		if (versions[i] >= rhs.versions[i]) {
			less = false;
			if (!greater) break;
		}
	}
}

Data_item::hash_type
Data_item::hash() const
{
	if (!_dirty) return _hash;

	// a 'const' function that needs to update the cached values
	Data_item *this_ = const_cast<Data_item *>(this);

	// recompute the hash
	MD5_hash md5;
	for (size_t i = 0; i < data_version::NUM_HOSTS; i++) {
#ifdef NBO
		uint32_t tmp32 = htobe32(_version.versions[i]);
#else
		uint32_t tmp32 = _version.versions[i];
#endif
		md5.update(&tmp32, 4);
	}
#ifdef NBO
	uint64_t tmp64 = htobe64(_version.timestamp);
#else
	uint64_t tmp64 = _version.timestamp;
#endif
	md5.update(&tmp64, 8);
	md5.update(_key);
	md5.update(_value);
	this_->_hash = md5.final64();
	this_->_dirty = false;
	return _hash;
}

data_version
Segment::set(const std::string &key, const std::string &value,
    uint8_t host_index)
{
	wrlock();
	map_type::iterator i = _store.find(key);
	if (i == _store.end()) {
		_store[key] = value_type(new Data_item(
		    key, value, host_index));
	} else {
		i->second->set(value, host_index);
	}
	_dirty = true;
	data_version vers = _store[key]->version();
	unlock();

	return vers;
}

// for anti-entropy (copy another version)
void
Segment::set(const std::string &key, const std::string &value,
    const data_version &version)
{
	wrlock();
	map_type::iterator i = _store.find(key);
	if (i == _store.end()) {
		_store[key] = value_type(new Data_item(
		    key, value, version));
	} else {
		i->second->set(value, version);
	}
	_dirty = true;
	unlock();
}

std::pair<std::string, data_version>
Segment::get(const std::string &key) throw (key_undefined)
{
	std::string value;
	data_version version;

	wrlock();
	map_type::iterator i = _store.find(key);
	if (i == _store.end()) {
		unlock();
		throw key_undefined();
	} else {
		value = i->second->value();
		version = i->second->version();
	}
	unlock();

	return std::make_pair(value, version);
}

Segment::hash_type
Segment::hash() const throw (need_lock)
{
	if (!_persistent_lock)
		throw need_lock();

	if (!_dirty)
		return _cached_hash;

	// a 'const' function that needs to update the cached values
	Segment *this_ = const_cast<Segment *>(this);

	// recompute hash
	MD5_hash md5;
	for (map_type::const_iterator i = _store.begin();
	    i != _store.end(); ++i) {
		// assuming below that hash_type is still 64 bits
		assert(sizeof(hash_type) == 8);
#ifdef NBO
		hash_type h = htobe64(i->second->hash());
#else
		hash_type h = i->second->hash();
#endif
		md5.update(&h, sizeof(h));
	}
	uint64_t digest = md5.final64();
	this_->_cached_hash = digest;
	this_->_dirty = false;

	return digest;
}

Segment_map::Segment_map(const struct sockaddr_in &owner_addr)
{
	std::vector<hash_type> hashes;
	segment_hashes(owner_addr, hashes);
	for (std::vector<hash_type>::iterator i = hashes.begin();
			i != hashes.end(); ++i){
		_map[*i] = value_type(new Segment);
#ifdef SEGMENT_DEBUG
		std::cerr << "we own: " << *i << "\n";
#endif
	}

	pthread_rwlock_init(&_lock, 0);
}

data_version
Segment_map::set(const hash_type &segment_hash, const std::string &key,
    const std::string &value, uint8_t host_index) throw (no_such_segment)
{
	rdlock();
	seg_map::iterator seg_iter = _map.find(segment_hash);
	if (seg_iter == _map.end()) {
		unlock();
#ifdef SEGMENT_DEBUG
		std::cerr << "No segment for hash " << segment_hash << "\n";
#endif
		throw no_such_segment();
	}
	value_type segment = seg_iter->second;
	data_version vers = segment->set(key, value, host_index);
	unlock();

	return vers;
}

void
Segment_map::set(const hash_type &segment_hash, const std::string &key,
    const std::string &value, const data_version &version)
    throw (no_such_segment)
{
	rdlock();
	seg_map::iterator seg_iter = _map.find(segment_hash);
	if (seg_iter == _map.end()) {
		unlock();
		throw no_such_segment();
	}
	value_type segment = seg_iter->second;
	segment->set(key, value, version);
	unlock();
}

std::pair<std::string, data_version>
Segment_map::get(const hash_type &segment_hash, const std::string &key) const
    throw (key_undefined, no_such_segment)
{
	rdlock();
	seg_map::const_iterator seg_iter = _map.find(segment_hash);
	if (seg_iter == _map.end()) {
		unlock();
#ifdef SEGMENT_DEBUG
		std::cerr << "No segment for hash " << segment_hash << "\n";
#endif
		throw no_such_segment();
	}
	value_type segment = seg_iter->second;
	std::pair<std::string, data_version> result;
	try {
		result = segment->get(key);
	} catch (...) {
		unlock();
		throw;
	}
	unlock();

	return result;
}

Segment *
Segment_map::segment_acquire(const hash_type &segment_hash)
    throw (no_such_segment)
{
	rdlock();
	seg_map::iterator iter = _map.find(segment_hash);
	if (iter == _map.end()) {
		std::map<hash_type, value_type>::iterator it;
#ifdef SEGMENT_DEBUG
		std::cout << "We are storing...\n";
		for (it = _map.begin(); it != _map.end(); it++){
			std::cout << (*it).first << "=>"
			    << (*it).second << "\n";
		}
#endif
		unlock();
#ifdef SEGMENT_DEBUG
		std::cerr << "No segment for hash " << segment_hash << "\n";
#endif
		throw no_such_segment();
	}
	Segment *segment = iter->second.get();
	segment->persistent_lock();
	unlock();

	return segment;
}

Segment_owner_map::Segment_owner_map(
    const std::vector<struct sockaddr_in> &addrs) :
	_addrs(addrs)
{
	using std::vector;

	std::sort(_addrs.begin(), _addrs.end(), Sockaddrin_cmp());

	for (vector<struct sockaddr_in>::iterator i = _addrs.begin();
	    i != _addrs.end(); ++i) {
		vector<uint64_t> hashes;
		segment_hashes(*i, hashes);
		for (uint8_t seg = 0; seg < hashes.size(); seg++) {
			_owners[hashes[seg]] = vector<struct sockaddr_in *>();
			// make the vector initially contain the
			// owner
			_owners[hashes[seg]].push_back(&*i);
		}
	}

	// construct the vectors in the '_owners' map
	// for each owner:
	for (iterator i = _owners.begin(); i != _owners.end(); ++i) {
		// set of owners in the vector so far
		std::set<struct sockaddr_in *> seen;
		seen.insert(i->second.front());

		// keep adding owners to the vector until all owners
		// are in it
		iterator j = i;
		while (seen.size() < _addrs.size()) {
			// advance j
			if (++j == _owners.end()) j = _owners.begin();

			// if addr not already in i's vector...
			struct sockaddr_in *addr = j->second.front();
			if (seen.insert(addr).second) {
				// ...then append it
				i->second.push_back(addr);
			}
		}
	}
}

Segment_owner_map::hash_type
Segment_owner_map::find_segment(const std::string &key) const
{
	MD5_hash md5;
	md5.update(key);
	return find_segment(md5.final64());
}
