/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef SEGMENT_HPP
#define SEGMENT_HPP

#include <netinet/in.h>
#include <stdint.h>

#ifdef SEGMENT_DEBUG
#	include <iostream>
#endif

#include <cassert>
#include <algorithm>
#include <map>
#include <string>
#include <vector>
#include <tr1/memory>

#include "clock.h"
#include "errors.hpp"

const uint8_t SEGMENTS_PER_NODE = 4;

void	segment_hashes(const struct sockaddr_in &addr,
	    std::vector<uint64_t> &hashes);

struct data_version {
	const static size_t NUM_HOSTS = 4;

	// the index of the host, according to ip-address/port ordering
	data_version() :
		timestamp(0)
	{
		std::fill(versions, versions + NUM_HOSTS, 0);
	}

	bool operator<(const data_version &rhs) const
	{
		bool greater;
		bool less;
		cmp(rhs, greater, less);
		return less || (!greater && timestamp < rhs.timestamp);
	}
	bool operator>(const data_version &rhs) const
	{
		return rhs < *this;
	}
	bool operator<=(const data_version &rhs) const
	{
		return !(*this > rhs);
	}
	bool operator>=(const data_version &rhs) const
	{
		return !(*this < rhs);
	}

	void update(size_t host_index)
	{
		assert(host_index < NUM_HOSTS);
		versions[host_index]++;
		timestamp = clock_monotonic();
	}

	// anti-entropy will fill this in manually
	uint32_t versions[NUM_HOSTS];
	uint64_t timestamp;

private:
	void cmp(const data_version &rhs, bool &greater, bool &less) const;
};

// a tuple of (version vector, key, value, hash(vector,key,value))
class Data_item {
public:
	typedef uint64_t	hash_type;

	Data_item(const std::string &key, const std::string &value,
	    size_t host_index) :
		_key(key),
		_value(value),
		_dirty(true)
	{
		_version.update(host_index);
	}

	// for anti-entropy (copy another version)
	Data_item(const std::string &key, const std::string &value,
	    const data_version &version) :
		_version(version),
		_key(key),
		_value(value),
		_dirty(true)
	{}

	const std::string &key() const
	{
		return _key;
	}
	const std::string &value() const
	{
		return _value;
	}
	data_version &version()
	{
		return _version;
	}
	const data_version &version() const
	{
		return _version;
	}
	hash_type hash() const;

	void set(const std::string &newval, size_t host_index)
	{
		_value = newval;
		_version.update(host_index);
		_dirty = true;
	}
	void set(const std::string &newval, const data_version &version)
	{
		_value = newval;
		_version = version;
		_dirty = true;
	}

private:
	// not copyable
	Data_item(const Data_item &) {}
	void operator=(const Data_item &) {}

	data_version 	_version;
	std::string	_key;
	std::string	_value;
	hash_type	_hash;  // hash of version, key, value
	bool		_dirty;
};

// maps keys to Data_item's
class Segment {
public:
	typedef uint64_t				hash_type;
	typedef std::string				key_type;
	typedef std::tr1::shared_ptr<Data_item>		value_type;
	typedef std::map<key_type, value_type>		map_type;

	Segment() :
		_dirty(true),
		_persistent_lock(false)
	{
		pthread_rwlock_init(&_lock, 0);
	}

	~Segment()
	{
		pthread_rwlock_destroy(&_lock);
	}

	// host_index: the host that initiated this change
	data_version set(const std::string &key, const std::string &value,
	    uint8_t host_index);

	// for anti-entropy (copy another version)
	void set(const std::string &key, const std::string &value,
	    const data_version &version);

	std::pair<std::string, data_version> get(const std::string &key)
	    throw (key_undefined);

	// only use the returned value while persistent_lock() is held
	map_type *map()
	{
		return &_store;
	}

	// return the current hash value
	// must call persistent_lock first
	hash_type hash() const throw (need_lock);

	// only use these for anti-entropy, so that hash()
	void persistent_lock()
	{
		wrlock();
		_persistent_lock = true;
	}
	void persistent_unlock()
	{
		_persistent_lock = false;
		// assume the hash needs to be recomputed
		_dirty = true;
		unlock();
	}

private:
	// not copyable
	Segment(const Segment &) {}
	void operator=(const Segment &) {}

	void rdlock() const
	{
		pthread_rwlock_rdlock(const_cast<pthread_rwlock_t *>(&_lock));
	}
	void wrlock()
	{
		pthread_rwlock_wrlock(&_lock);
	}
	void unlock() const
	{
		pthread_rwlock_unlock(const_cast<pthread_rwlock_t *>(&_lock));
	}

	map_type		_store;
	hash_type		_cached_hash;
	bool			_dirty;
	pthread_rwlock_t	_lock;
	bool			_persistent_lock;
};

// frontend for setting keys, getting values, and referencing Segment's
// for anti-entropy purposes
class Segment_map {
public:
	typedef uint64_t							hash_type;
	typedef std::tr1::shared_ptr<Segment>		value_type;
	typedef std::map<hash_type, value_type>		seg_map;

	// start a segment map initially containing just segments of the owner
	explicit Segment_map(const struct sockaddr_in &owner_addr);
	~Segment_map()
	{
		pthread_rwlock_destroy(&_lock);
	}

	// host_index is the index number (under some total ordering of the
	// hosts, probably by ip-addr, then port)
	data_version set(const hash_type &segment_hash,
	    const std::string &key, const std::string &value,
	    uint8_t host_index)
	    throw (no_such_segment);

	void set(const hash_type &segment_hash,
	    const std::string &key, const std::string &value,
	    const data_version &version)
	    throw (no_such_segment);

	//The segment_hash is the output of find_segment.
	std::pair<std::string, data_version> get(
	    const hash_type &segment_hash, const std::string &key) const
	    throw (key_undefined, no_such_segment);

	// anti-entropy only; the persistent_lock() function will already
	// have been called, so you only get the map and hash of the Segment;
	// you *must* call segment_release on completion
	Segment *segment_acquire(const hash_type &segment_hash)
	    throw (no_such_segment);

	void segment_release(Segment *segment)
	{
		segment->persistent_unlock();
	}

	void link(const hash_type &segment_hash)
	{
		wrlock();
		if (_map.find(segment_hash) == _map.end()) {
			_map[segment_hash] = value_type(new Segment);
#ifdef SEGMENT_DEBUG
			std::cerr << "added segment " << segment_hash << '\n';
#endif
		}
		unlock();
	}

	std::tr1::shared_ptr<Segment> unlink(const hash_type &segment_hash)
	{
		wrlock();
		seg_map::iterator i = _map.find(segment_hash);
		value_type segment;
		if (i != _map.end()) {
			segment = i->second;
			segment->persistent_lock();
			_map.erase(i);
		}
		unlock();

		return segment;
	}

private:
	// not copyable
	Segment_map(const Segment_map &) {}
	void operator=(const Segment_map &) {}

	void rdlock() const
	{
		pthread_rwlock_rdlock(const_cast<pthread_rwlock_t *>(&_lock));
	}
	void wrlock()
	{
		pthread_rwlock_wrlock(&_lock);
	}
	void unlock() const
	{
		pthread_rwlock_unlock(const_cast<pthread_rwlock_t *>(&_lock));
	}

	seg_map			_map;
	pthread_rwlock_t	_lock;
};

// maps key hashes to segment hashes and indicates which nodes are backups
class Segment_owner_map {
public:
	typedef uint64_t			hash_type;
	typedef std::map<hash_type, std::vector<struct sockaddr_in *> >
						hash_owner_map;

	typedef hash_owner_map::iterator	iterator;
	typedef hash_owner_map::const_iterator	const_iterator;

	explicit Segment_owner_map(
	    const std::vector<struct sockaddr_in> &addrs);

	// find hash of segment (used as a lookup for other mapping functions)
	hash_type find_segment(const hash_type &hash) const
	{
		Segment_owner_map *this_ =
		    const_cast<Segment_owner_map *>(this);
		iterator iter = this_->first_node(hash);
		return iter->first;
	}
	hash_type find_segment(const std::string &key) const;

	// return a vector (same length as number of nodes) of the owner
	// node's and the addresses of the N-1 successors
	const std::vector<struct sockaddr_in *> &successors(
	    const hash_type &segment_hash) const
	{
		const_iterator iter = _owners.find(segment_hash);
		if (iter == _owners.end()) throw server_error();
		return iter->second;
	}

	size_t host_index(const struct sockaddr_in &addr) const
	{
		for (size_t i = 0; i < _addrs.size(); i++)
			if (_addrs[i].sin_port == addr.sin_port &&
			    _addrs[i].sin_addr.s_addr == addr.sin_addr.s_addr)
				return i;
		return _addrs.size();
	}

private:
	// not copyable
	Segment_owner_map(const Segment_owner_map &) {}
	void operator=(const Segment_owner_map &) {}

	iterator first_node(const hash_type &hash)
	{
		// first element with key > 'hash'
		iterator i = _owners.upper_bound(hash);

		// if 'hash' larger than largest point
		if (i == _owners.end()) return _owners.begin();

		// check if there is a node with key == 'hash'
		iterator j = predecessor(i);
		if (j->first == hash) return j;

		return i;
	}
	iterator predecessor(iterator i)
	{
		if (i == _owners.begin()) i = _owners.end();
		return --i;
	}
	iterator successor(iterator i)
	{
		if (++i == _owners.end()) return _owners.begin();
		return i;
	}

	hash_owner_map			_owners;
	std::vector<struct sockaddr_in>	_addrs; // ordered
};

#endif
