/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <arpa/inet.h>
#include <netinet/in.h>
#include <signal.h>

#include <cassert>
#include <iostream>

#include <getopt.h>
#include <stdint.h>

#include "ae_queue.hpp"
#include "anti_entropy.hpp"
#include "heartbeat.hpp"
#include "parse.hpp"
#include "server_listener.hpp"
#include "server_queue.hpp"
#include "server_table.hpp"
#include "server_worker.hpp"

const char *prog;

// anon namespace
namespace {

const unsigned NUM_WORKERS = 10;
const unsigned NUM_ANTI_ENTROPY = 4;

// just copied from clientd.cpp
struct thread_owner {
	thread_owner(pthread_t thread=0) : thread(thread) {}
	~thread_owner()
	{
		if (thread) pthread_kill(thread, SIGKILL);
	}

	pthread_t release()
	{
		pthread_t old = thread;
		thread = 0;
		return old;
	}

	pthread_t thread;

private:
	// non-copyable
	thread_owner(const thread_owner &) {}
	void operator=(const thread_owner &) {}
};

/******************************************************************************
* server_setup
* takes:	&host_addrs  vector of sockaddr_ins for the servers
*
* does:		creates:
*				listener thread
*				worker thread workpool
*				antientropy listener, active, responders
*				heartbeat maker, monitor
******************************************************************************/
void
server_setup(std::vector<struct sockaddr_in> &host_addrs)
    throw (general_error)
{
	using std::vector;
	using std::tr1::shared_ptr;

	struct sockaddr_in self_addr = host_addrs.front();

	// shared data structures
	AE_queue ae_queue;
	AE_segment_queue ae_seg_queue(self_addr);
	Server_queue request_queue;
	Server_table host_table(self_addr, host_addrs);
	Segment_owner_map owner_map(host_addrs);
	Segment_map seg_map(self_addr);

	int errval;
	int listen_socket = server_start_listening(self_addr);

	vector<pthread_t> all_threads;
	vector<shared_ptr<thread_owner> > all_owners;

	// NOTE: for each thread below, there is an argument structure (either
	// on the stack or added to some vector of shared_ptr).  After the
	// thread is created, the thread and thread owner are added to
	// 'all_threads' and 'all_owners' so they can be waited on later.

	// create listener
	pthread_t listener;
	server_listener_args server_listener_args(ae_queue, request_queue,
	    listen_socket);
	if ((errval = pthread_create(&listener, 0,
	    server_listener, &server_listener_args)))
		throw unix_error("creating listener thread", errval);
	all_threads.push_back(listener);
	all_owners.push_back(shared_ptr<thread_owner>(
	    new thread_owner(listener)));

	// create workers
	vector<shared_ptr<server_worker_args> > worker_args;
	for (unsigned i = 0; i < NUM_WORKERS; i++) {
		pthread_t worker;
		shared_ptr<server_worker_args> args(new server_worker_args(
				host_table, request_queue, i, owner_map, seg_map));

		if ((errval = pthread_create(&worker, 0, server_worker,
		    args.get())))
			throw unix_error("creating worker thread", errval);

		worker_args.push_back(args);
		all_threads.push_back(worker);
		all_owners.push_back(shared_ptr<thread_owner>(
		    new thread_owner(worker)));
	}

	// create anti-entropy responder workers
	vector<shared_ptr<ae_respond_args> > ae_responder_args;
	vector<shared_ptr<thread_owner> > ae_responder_owners;
	for (unsigned i = 0; i < NUM_ANTI_ENTROPY; i++) {
		pthread_t thread;
		shared_ptr<ae_respond_args> args(
		    new ae_respond_args(ae_queue, seg_map));

		if ((errval = pthread_create(&thread, 0, ae_respond,
		    args.get())))
			throw unix_error("creating anti-entropy responder",
			    errval);

		ae_responder_args.push_back(args);
		all_threads.push_back(thread);
		all_owners.push_back(shared_ptr<thread_owner>(
		    new thread_owner(thread)));
	}

	// create anti-entropy active workers
	vector<shared_ptr<ae_active_args> > ae_active_arg_list;
	vector<shared_ptr<thread_owner> > ae_active_owners;
	for (unsigned i = 0; i < NUM_ANTI_ENTROPY; i++) {
		pthread_t thread;
		shared_ptr<ae_active_args> args(new ae_active_args(
		    ae_seg_queue, seg_map, owner_map));

		if ((errval = pthread_create(&thread, 0, ae_active,
		    args.get())))
			throw unix_error("creating anti-entropy active",
			    errval);

		ae_active_arg_list.push_back(args);
		all_threads.push_back(thread);
		all_owners.push_back(shared_ptr<thread_owner>(
		    new thread_owner(thread)));
	}

	// create heartbeat threads
	pthread_t heartbeat_monitor_thread;
	heartbeat_monitor_args heartbeat_monitor_args(host_table);
	if ((errval = pthread_create(&heartbeat_monitor_thread, 0,
	    heartbeat_monitor, &heartbeat_monitor_args)))
		throw unix_error("creating heartbeat monitor thread", errval);
	all_threads.push_back(heartbeat_monitor_thread);
	all_owners.push_back(shared_ptr<thread_owner>(
	    new thread_owner(heartbeat_monitor_thread)));

	pthread_t heartbeat_maker_thread;
	heartbeat_maker_args heartbeat_maker_args(host_table);
	if ((errval = pthread_create(&heartbeat_maker_thread, 0,
	    heartbeat_maker, &heartbeat_maker_args)))
		throw unix_error("creating heartbeat maker thread", errval);
	all_threads.push_back(heartbeat_maker_thread);
	all_owners.push_back(shared_ptr<thread_owner>(
	    new thread_owner(heartbeat_maker_thread)));

	std::cerr << ">>> server listening for requests\n";

	// wait for each of the whole mess of threads, so as to never exit
	for (size_t i = 0; i < all_threads.size(); i++) {
		pthread_join(all_threads[i], 0);
		all_owners[i]->release();
	}

	throw server_error("all threads exited??");
}

/******************************************************************************
* usage
*
* does:		prints usage information
******************************************************************************/
void
usage()
{
	std::cout << "usage: " << prog << '\n';
}

} // end anon namespace

int
main(int argc, char **argv)
{
	using std::cerr;
	using std::cout;
	using std::pair;
	using std::string;
	using std::vector;

	string	config_path = SERVER_CONFIG;

	prog = *argv;

	// we'll print errors ourselves
	opterr = 0;

	// use large values otherwise not returned by getopt_long()
	enum { OPT_HELP = 256 };
	struct option longopts[] = {
		{ "help", 0, 0, OPT_HELP },
		{ 0, 0, 0, 0 }
	};

	for (;;) {
		// no options now ... just in case, though
		int opt = getopt_long(argc, argv, ":h", longopts, 0);
		if (opt == -1) break;

		switch (opt) {
		case OPT_HELP:
		case 'h':
			usage();
			return 0;
		case ':':
			cerr << prog << ": missing argument for -"
			    << static_cast<char>(optopt) << '\n';
			return 1;
		case '?':
			cerr << prog << ": unrecognized option -"
			    << static_cast<char>(optopt) << '\n';
			return 1;
		default:
			assert(0);
		}
	}

	vector<pair<string, uint16_t> > hosts;
	vector<struct sockaddr_in> host_addrs;
	try {
		// read and parse config
		config_parse(config_path, hosts);
		resolve_hosts(hosts, host_addrs);

		// setup threads, then wait forever
		server_setup(host_addrs);
	} catch (const general_error &e) {
		cerr << prog << ": " << e.what() << '\n';
		return 1;
	}

	return 0;
}
