/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <sys/socket.h>

#include <cassert>
#include <cerrno>
#include <cstring>
#include <iostream>
#include <tr1/memory>

#include "ae_queue.hpp"
#include "server_listener.hpp"

const int SERVER_BACKLOG = 20;

extern char *prog;

// start listening, return socket
int
server_start_listening(const struct sockaddr_in &addr) throw (unix_error)
{
	int sockfd;

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		throw unix_error("creating TCP listen socket");

	int rebind = 1;
	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &rebind,
	    sizeof(rebind)) == -1) {
		long error = errno;
		close(sockfd);
		throw unix_error("reusing the address", error);
	}

	if (bind(sockfd, reinterpret_cast<const struct sockaddr *>(&addr),
	    sizeof(addr)) == -1) {
		long error = errno;
		close(sockfd);
		throw unix_error("connecting server TCP listen socket", error);
	}

	if (listen(sockfd, SERVER_BACKLOG) == -1) {
		long error = errno;
		close(sockfd);
		throw unix_error("TCP listen failed", error);
	}

	return sockfd;
}

void *
server_listener(void *voidarg)
{
	using std::cerr;
	using std::tr1::shared_ptr;

	server_listener_args *args = reinterpret_cast<server_listener_args *>(
	    voidarg);

	// copy arguments
	AE_queue	&ae_queue = args->ae_queue;
	Server_queue	&server_queue = args->server_queue;
	int		sock = args->sock;

	for (;;) {
		int fd;
		struct sockaddr_in addr;
		socklen_t addrlen = sizeof(addr);

		if ((fd = accept(sock,
		    reinterpret_cast<struct sockaddr *>(&addr), &addrlen))
		    == -1) {
			cerr << prog << "(listener): accept failed: "
			    << strerror(errno) << '\n';
			continue;
		}

		// read first three bytes to route the request
		char	buf[4];
		uint8_t	size = 0;
		while (size < 3) {
			ssize_t bytes;
			if ((bytes = read(fd, buf + size, 3 - size)) == -1) {
				if (errno == EINTR) continue;
				cerr << prog << "(listener): read failed: "
				    << strerror(errno) << '\n';
				close(fd);
				fd = -1;
				break;
			} else if (!bytes) {
				cerr << prog << "(listener): premature EOF\n";
				close(fd);
				fd = -1;
				break;
			}
			size += bytes;
		}
		if (fd == -1) continue;
		buf[3] = '\0';

		std::string request = buf;
		bool is_http = request == "GET";
		bool is_ae = request == "ANT";

#ifdef SERVER_DEBUG
		std::cerr << "listener: received "
		    << (is_http ? "HTTP" : is_ae ? "AE" : "????")
		    << " request\n";
#endif

		if (is_http) {
			server_queue.push(shared_ptr<Server_connection>(
			    new Server_connection(fd)));
		} else if (is_ae) {
			ae_queue.push(shared_ptr<AE_connection>(
			    new AE_connection(fd)));
		} else {
			assert(0);
		}
	}

	return 0;
}
