/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef SERVER_QUEUE_HPP
#define SERVER_QUEUE_HPP

#include <pthread.h>
#include <unistd.h>

#include <cassert>
#include <queue>
#include <string>
#include <tr1/memory>
#include <stdint.h>

struct Server_connection {
	explicit Server_connection(int socket) :
		socket(socket)
	{}
	~Server_connection()
	{
		close(socket);
	}

	int		socket;

private:
	// non-copyable
	Server_connection(const Server_connection &) {}
	void operator=(const Server_connection &) {}
};

// a synchronized queue for client requests
class Server_queue {
public:
	Server_queue()
	{
		pthread_cond_init(&cond, 0);
		pthread_mutex_init(&mutex, 0);
	}
	~Server_queue()
	{
		pthread_cond_destroy(&cond);
		pthread_mutex_destroy(&mutex);
	}

	void push(std::tr1::shared_ptr<Server_connection> cx)
	{
		// XXX presently the queue size is unbounded
		pthread_mutex_lock(&mutex);

		queue.push(cx);

		pthread_cond_signal(&cond);
		pthread_mutex_unlock(&mutex);
	}

	// pop without waiting, returning a null pointer if empty
	std::tr1::shared_ptr<Server_connection> pop_nowait()
	{
		pthread_mutex_lock(&mutex);

		std::tr1::shared_ptr<Server_connection> old = queue.front();
		queue.pop();

		if (!queue.empty())
			pthread_cond_signal(&cond);
		pthread_mutex_unlock(&mutex);

		return old;
	}

	// pop with waiting
	std::tr1::shared_ptr<Server_connection> pop()
	{
		pthread_mutex_lock(&mutex);
		while (queue.empty())
			pthread_cond_wait(&cond, &mutex);

		std::tr1::shared_ptr<Server_connection> old = queue.front();
		queue.pop();

		if (!queue.empty())
			pthread_cond_signal(&cond);
		pthread_mutex_unlock(&mutex);
		return old;
	}

private:
	// not copyable
	Server_queue(const Server_queue &) {}
	void operator=(const Server_queue &) {}

	std::queue<std::tr1::shared_ptr<Server_connection> > queue;
	pthread_cond_t		cond;
	pthread_mutex_t		mutex;
};

#endif
