/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef SERVER_TABLE_HPP
#define SERVER_TABLE_HPP

#include <netinet/in.h>
#include <arpa/inet.h>

#include <cassert>
#include <vector>

//#ifdef SERVER_DEBUG
#	include <iostream>
//#endif

#include "clock.h"
#include "segment.hpp"
#include "util.hpp"

// used to find good nodes; it should be updated by threads that find an
// alive node; the heartbeat monitor will update the table also; the heartbeat
// transmitter will ALSO touch this
class Server_table {
public:
	typedef uint16_t	hash_type;

	Server_table(const struct sockaddr_in &self,
	    const std::vector<struct sockaddr_in> &host_addrs)
	{
		pthread_rwlock_init(&table_lock, 0);
		table_vector.reserve(host_addrs.size());
		//Memory bugs?
		std::vector<struct sockaddr_in>::const_iterator it;
		for (it=host_addrs.begin(); it != host_addrs.end(); it++){
			if (compare_sockaddr_in(self, *it)){
				table_vector.push_back(node_properties(
				    *it, true));
				self_pointer = &(table_vector.back().addr);
				table_vector.back().up = true;
			}
			else {
				table_vector.push_back(node_properties(
				    *it, false));
			}
		}
		assert(self_pointer);
	}

	void touch(const struct sockaddr_in &addr) 
	{
		update_timestamp(addr, clock_monotonic());
	}

	//So sometimes it is probably better to update with a specific time; udp
	// can take a REALLY long time (well not in this case probably)
	void update_timestamp(const struct sockaddr_in &addr, uint64_t timestamp) {
		wrlock();
		struct node_properties * node = find_entry(addr);
		if (node->last_received < timestamp){
			node->last_received = timestamp;
			if (!node->up){
//#ifdef SERVER_DEBUG
				std::cout << addr_to_str(addr)
				    << " is now up\n";
//#endif
				node->up = true;
			}
		}
		unlock();
	}

	const struct sockaddr_in &get_self() const{
		return *self_pointer;
	}

	uint64_t get_last_timestamp(const struct sockaddr_in &addr){
		rdlock();
		uint64_t retvalue =  find_entry(addr)->last_received;
		unlock();
		return retvalue;
	}

	void update_up(const struct sockaddr_in &addr, bool up){
		wrlock();
		struct node_properties * node = find_entry(addr);
//#ifdef SERVER_DEBUG
		if (node->up != up){
			std::cout << addr_to_str(addr);
			std::cout << " is now " << (up?"up":"down") << "\n";
		}
//#endif
		node->up = up;
		unlock();
	}

	bool is_up(const struct sockaddr_in &addr) const{
		rdlock();
		bool ret = find_entry(addr)->up;
		unlock();
		return ret;
	}

	//Not sure where this really belongs, so I'm sticking it here.
	// We often want to tell if two sockaddr_in s refer to the same server
	static int compare_sockaddr_in(const struct sockaddr_in &a,
	    const struct sockaddr_in &b){
		return (a.sin_addr.s_addr ==  b.sin_addr.s_addr && a.sin_port == b.sin_port);
	}
	
	std::vector<const struct sockaddr_in *> get_nodes() const {
		rdlock();
#ifdef SERVER_DEBUG
		std::cout << "Table vector size"
		    << table_vector.size() << "\n";
#endif
		std::vector<struct node_properties>::const_iterator it;
		std::vector<const struct sockaddr_in * > nodes;
		for (it=table_vector.begin(); it < table_vector.end(); it++){
			nodes.push_back(&((*it).addr));
		}
		unlock();
		return nodes;
	}

private:
	// not copyable
	Server_table(const Server_table &) {}
	void operator=(const Server_table &) {}

	struct node_properties {
		node_properties(const struct sockaddr_in &addr, bool self) :
			addr(addr),
			last_received(0),
			up(0),
			self(self)
		{}

		struct sockaddr_in	addr;
		uint64_t		last_received;
		bool			up;
		bool			self;
	};

	pthread_rwlock_t table_lock;

	typedef std::vector<std::tr1::shared_ptr<node_properties> >
	    addr_vector;

	std::map<hash_type, addr_vector::iterator>	segment_points;
	
	std::vector<struct node_properties> table_vector;
	struct sockaddr_in * self_pointer; //set to point to myself
	
	const struct node_properties * find_entry(
	    const struct sockaddr_in &addr) const{
#ifdef SERVER_DEBUG
		std::cout << "searching for " << addr_to_str(addr) << '\n';
#endif
		std::vector<struct node_properties>::const_iterator it;
		for (it=table_vector.begin(); it != table_vector.end(); ++it){
			if (compare_sockaddr_in(it->addr, addr)){
				return &(*it); // =(
			}
#ifdef SERVER_DEBUG
			std::cout << "current: " << addr_to_str(addr) << '\n';
#endif
		}
		assert(0);
	}
	//  non-const version
	struct node_properties *find_entry(const struct sockaddr_in &addr) {
		return const_cast<struct node_properties *>(
		    const_cast<const Server_table *>(this)->find_entry(addr));
	}

	//lifted from segment.hpp
	void rdlock() const
	{
		pthread_rwlock_rdlock(const_cast<pthread_rwlock_t *>(&table_lock));
	}
	void wrlock()
	{
		pthread_rwlock_wrlock(&table_lock);
	}
	void unlock() const
	{
		pthread_rwlock_unlock(const_cast<pthread_rwlock_t *>(&table_lock));
	}
};


#endif
