/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>, Jeffery Roller <roller-j@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <unistd.h>

#include <cerrno>
#include <climits>
#include <cstring>
#include <iostream>
#include <sstream>

#include "server_worker.hpp"
#include "util.hpp"

#define BACKLOG_SIZE 100

extern const char *prog;

namespace {

std::string
worker_name(unsigned num)
{
	std::ostringstream out(prog);
	out << "(worker " << num << ')';
	return out.str();
}

inline bool
operator==(const struct sockaddr_in &a, const struct sockaddr_in &b)
{
	return a.sin_family == b.sin_family &&
	    a.sin_port == b.sin_port &&
	    a.sin_addr.s_addr == b.sin_addr.s_addr;
}

// stolen from clientd_worker.cpp
std::string
hostname() throw (unix_error)
{
	char buf[HOST_NAME_MAX + 1];
	if (gethostname(buf, HOST_NAME_MAX) == -1)
		throw unix_error("gethostname() failed");
	buf[HOST_NAME_MAX] = '\0';
	return std::string(buf);
}

std::tr1::shared_ptr<Http_response>
forward_request(const Http_request &similar, const struct sockaddr_in &addr,
    bool request_forward, bool include_version, const data_version &version)
{
	std::ostringstream out;

	out << "GET [" << similar.key() << ']';
	if (similar.have_value())
		out << "=[" << similar.value() << ']';
	out << " HTTP/1.1\r\n"
	    << "Host: " << hostname() << "\r\n"
	    << "User-Agent: " << SERVER_USER_AGENT << "\r\n";
	if (!request_forward)
		out << "Pragma: no_forwarding="
		    << (request_forward ? "false" : "true") << "\r\n";
	if (include_version) {
		out << "Pragma: version=";
		for (size_t i = 0; i < data_version::NUM_HOSTS; i++) {
			out << version.versions[i] << ',';
		}
		out << version.timestamp << "\r\n";
	}
	out << "\r\n";

	std::tr1::shared_ptr<Http_response> response;
	int sock = -1;
#ifdef SERVER_DEBUG
	std::cerr << "out is " << out;
	std::cerr << " and address is " << addr_to_str(addr) << "\n";
#endif
	try {
		sock = send_to_addr(out.str(), &addr);
	} catch (const general_error &e) {
		std::cerr << "error forwarding request: " << e.what() << '\n';
	}
	if (sock != -1) {
		try {
			response.reset(new Http_response("", sock));
		} catch (const general_error &e) {
			close(sock);
			std::cerr << "error forwarding request: "
			    << e.what() << '\n';
		}
		close(sock);
#ifdef SERVER_DEBUG
		std::cerr << "received forward response\n";
#endif
	}

	return response;
}

std::string
client_response(const std::string &key, bool have_value,
    const std::string &value, const data_version &new_version,
    const Server_table &table)
{
	std::ostringstream out;
	// brackets, newline, equal: 5
	size_t content_length = key.size() + 5 +
	    (have_value ? value.size() + 2 : 3);
	std::vector<const struct sockaddr_in *> nodes = table.get_nodes();

	out << "HTTP/1.0 200 OK\r\n";
	out << "Server: " << hostname() << "\r\n";
	out << "Content-Length: " << content_length << "\r\n";
	out << "Pragma: version=";
	for (size_t i = 0; i < data_version::NUM_HOSTS; i++) {
		out << new_version.versions[i] << ',';
	}
	out << new_version.timestamp << "\r\n";
	out << "Pragma: servers=";
	for (size_t i = 0; i < nodes.size(); i++) {
		const struct sockaddr_in &addr = *nodes[i];
		if (!table.is_up(addr)) continue;
		if (i) out << ',';
		out << addr_to_str(addr);
	}
	out << "\r\n";
	out << "Content-Type: text/plain\r\n";
	out << "\r\n"; //after this comes the message-body
	out << '[' << key << "]=";
	if (have_value)
		out << '[' << value << ']';
	else
		out << "NIL";
	out << "\r\n";
	return out.str();
}

// when naught can be done about it
std::string
err_response(const Server_table &table)
{
	std::ostringstream out;
	std::vector<const struct sockaddr_in *> nodes = table.get_nodes();

	// maybe a better error number?
	out << "HTTP/1.0 500 Error\r\n";
	out << "Server: " << hostname() << "\r\n";
	out << "Content-Length: 0\r\n";
	out << "Pragma: servers=";
	for (size_t i = 0; i < nodes.size(); i++) {
		const struct sockaddr_in &addr = *nodes[i];
		if (!table.is_up(addr)) continue;
		if (i) out << ',';
		out << addr_to_str(addr);
	}
	out << "\r\n";
	out << "Content-Type: text/plain\r\n";
	out << "\r\n";
	return out.str();
}

} // end anon namespace

void *
server_worker(void *voidarg)
{
	using std::cerr;
	using std::ostringstream;
	using std::string;
	using std::tr1::shared_ptr;

	typedef uint64_t			hash_type;
	server_worker_args *args = reinterpret_cast<server_worker_args *>(
	    voidarg);

	Server_queue	&queue = args->queue;
	Server_table	&server_table = args->servers;
	unsigned	worker_number = args->worker_number;
	Segment_owner_map &segment_owner_map = args->segment_owner_map;
	Segment_map	&segment_map = args->segment_map;

	string name = worker_name(worker_number);
	struct sockaddr_in self = server_table.get_self();
	int self_index = segment_owner_map.host_index(self);

	for (;;) {
		shared_ptr<Server_connection> cx = queue.pop(); //wait until something hits the queue
		int fd = cx->socket;

		shared_ptr<Http_request> req;
		try {
			req.reset(new Http_request("GET", fd));
		} catch (const general_error &e) {
			cerr << name << ": " << e.what() << '\n';
			continue;
		}

		bool is_put = req->have_value();
		bool is_server = req->client() == SERVER_USER_AGENT;
		//bool is_ourclient = req->client() == USER_AGENT;

#ifdef SERVER_DEBUG
		if (req->have_value()) {
			cerr << "worker" << worker_number
			    << ": received request put(" << req->key()
			    << ',' << req->value() << ")\n";
		} else
			cerr << "worker" << worker_number
			    << ": received request get(" << req->key()
			    << ")\n";
#endif

		// check who handles req:
		hash_type seg_hash = segment_owner_map.find_segment(
		    req->key());
		std::vector<struct sockaddr_in *> storing_nodes =
		    segment_owner_map.successors(seg_hash);

		struct sockaddr_in *primary = 0;
		struct sockaddr_in *secondary = 0;

#ifdef FAILOVER
		size_t considered = storing_nodes.size();
#else
		size_t considered = storing_nodes.size() < 2 ?
		    storing_nodes.size() : 2;
#endif
		
		for (size_t i = 0; i < considered; i++){
			if (server_table.is_up(*storing_nodes[i])) {
				if (primary) {
#ifdef SERVER_DEBUG
					cerr << "Setting secondary\n";
#endif
					secondary = storing_nodes[i];
					break;
				}
#ifdef SERVER_DEBUG
				cerr << "Setting primary\n";
#endif
				primary = storing_nodes[i];
			}
		}

#ifdef SERVER_DEBUG
		if (primary)
			cerr << "primary: " << addr_to_str(*primary) << '\n';
		if (secondary)
			cerr << "secondary: " << addr_to_str(*secondary)
			    << '\n';
#endif
		

		bool am_primary = primary && self == *primary;
		bool am_secondary = secondary && self == *secondary;
		bool had_value = false;
		std::pair<string, data_version> old_value;
		data_version new_version;
		shared_ptr<Http_response> response;

		bool read_value = (is_put && am_primary) ||
		    (!is_put && (am_primary || am_secondary));
		if (read_value) {
			try {
				old_value = segment_map.get(seg_hash,
				    req->key());
				had_value = true;
			} catch (key_undefined) {
#ifdef SERVER_DEBUG
				std::cerr << "requested key undefined\n";
#endif
			} catch (no_such_segment) {
#ifdef SERVER_DEBUG
				std::cerr << "segment did not exist\n";
#endif
			}
		}

		bool okay = false;

#ifdef SERVER_DEBUG
		if (am_primary)
			cerr << "I think I am the primary\n";
		if (am_secondary)
			cerr << "I think I am the secondary\n";
#endif
		
		if (is_put && am_primary) {
			// write locally
#ifdef SERVER_DEBUG
			cerr << "primary, write\n";
#endif
			try {
				segment_map.link(seg_hash);
				new_version = segment_map.set(seg_hash,
				    req->key(), req->value(), self_index);
			} catch (no_such_segment) {
				assert(0);
			}

			// forward to secondary
			if (secondary && !req->no_forwarding()) {
				response = forward_request(*req,
				    *secondary, false, true, new_version);
			}

			okay = true;
		} else if (is_put && am_secondary && req->no_forwarding()) {
			// write locally
#ifdef SERVER_DEBUG
			cerr << "secondary, write from primary\n";
#endif
			new_version = req->version();
			try {
				segment_map.link(seg_hash);
				segment_map.set(seg_hash,
				    req->key(), req->value(), req->version());
			} catch (no_such_segment) {
				assert(0);
			}
#ifdef SERVER_DEBUG
			cerr << "wrote " << req->key() << "=" << req->value()
			    << "\n";
#endif

			okay = true;
		} else if (is_put && am_secondary) {
			// forward to primary, no hops
#ifdef SERVER_DEBUG
			cerr << "secondary, write from client\n";
#endif
			response = forward_request(*req,
			    *primary, false, false, data_version());

			if (response.get()) {
				// return primary's values to client
				had_value = response->have_value();
				old_value = std::make_pair(response->value(),
				    response->version());
			} else {
				// read our own value
				try {
					old_value = segment_map.get(seg_hash,
					    req->key());
					had_value = true;
				} catch (key_undefined) {
				} catch (no_such_segment) {
#ifdef SERVER_DEBUG
					cerr << "!! lost own (secondary) "
					    "segment !!\n";
#endif
				}
			}

			// write value
			try {
				segment_map.link(seg_hash);
				if (response) {
					new_version = response->version();
					segment_map.set(
					    seg_hash, req->key(),
					    req->value(), new_version);
				} else
					new_version = segment_map.set(
					    seg_hash, req->key(),
					    req->value(), self_index);
			} catch (no_such_segment) {
				assert(0);
			}

			okay = true;
		} else if (is_put && !is_server && primary) {
			// forward to primary, no hops
#ifdef SERVER_DEBUG
			cerr << "other, write from client\n";
#endif
			response = forward_request(*req,
			    *primary, true, false, data_version());
			if (!response.get() && secondary) {
				response = forward_request(*req,
				    *secondary, true, false, data_version());
			}
			if (response.get()) {
				had_value = response->have_value();
				old_value = std::make_pair(response->value(),
				    response->version());

				okay = true;
			}
		} else if (!is_put && (am_primary || am_secondary)) {
#ifdef SERVER_DEBUG
			cerr << "primary/secondary, read\n";
#endif
			okay = true;
		} else if (!is_put && !is_server && primary) {
#ifdef SERVER_DEBUG
			cerr << "other, read\n";
#endif
			// try primary
			response = forward_request(*req,
			    *primary, true, false, data_version());
			if (secondary && !response.get())
				// try secondary
				response = forward_request(*req,
				    *primary, true, false, data_version());
			if (response.get()) {
				had_value = response->have_value();
				old_value = std::make_pair(response->value(),
				    response->version());
				okay = true;
			}
		} else {
#ifdef SERVER_DEBUG
			cerr << "BAD BAD BAD\n";
#endif
			// bad request from server; reject
		}

		std::string response_str;
		if (okay) {
//#ifdef SERVER_DEBUG
			cerr << "worker" << worker_number
			    << ": responding with data\n";
//#endif
			response_str = client_response(req->key(), had_value,
			    old_value.first, new_version, server_table);
		} else {
//#ifdef SERVER_DEBUG
			cerr << "worker" << worker_number
			    << ": responding with error\n";
//#endif
			response_str = err_response(server_table);
		}

		try {
			send_put_server(response_str, fd);
		} catch (const general_error &e) {
			cerr << name << ": " << e.what() << '\n';
		}

		/*
		bool done = false;
		while (!done){
			try {
				handle_request(req.get(), server_table, name,
				    segment_owner_map, segment_map, fd);
				done = true;
			} catch (request_restart) {
				done=false;
			}
		}
		*/
	}

	return 0;
}
